-- 000-DataTypes.sql
-- ---------------------------------------
-- Author:  Douglas Creager (dcreager@alum.mit.edu)
-- Copyright 1999-2002 Open Microscopy Environment (OME)
-- This file is part of OME.
-- ---------------------------------------
-- 
--     OME is free software; you can redistribute it and/or modify
--     it under the terms of the GNU General Public License as published by
--     the Free Software Foundation; either version 2 of the License, or
--     (at your option) any later version.
-- 
--     OME is distributed in the hope that it will be useful,
--     but WITHOUT ANY WARRANTY; without even the implied warranty of
--     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--     GNU General Public License for more details.
-- 
--     You should have received a copy of the GNU General Public License
--     along with OME; if not, write to the Free Software
--     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-- 


--
-- 24 September 2002 (DC)
-- Some new tables for the analysis subsystem

CREATE SEQUENCE LOOKUP_TABLE_SEQ;
CREATE TABLE LOOKUP_TABLES (
  LOOKUP_TABLE_ID  OID DEFAULT nextval('LOOKUP_TABLE_SEQ') PRIMARY KEY,
  NAME             VARCHAR(64) NOT NULL,
  DESCRIPTION      TEXT
);
CREATE INDEX LOOKUP_TABLES_NAME ON LOOKUP_TABLES(NAME);
COMMENT ON TABLE LOOKUP_TABLES IS
'This table and LOOKUP_TABLE_ENTRIES make up a mechanism to restrict \
certain module inputs to a list of values.  This table identifies the \
list, and LOOKUP_TABLE_ENTRIES identifies the values for that list.';

CREATE SEQUENCE LOOKUP_TABLE_ENTRY_SEQ;
CREATE TABLE LOOKUP_TABLE_ENTRIES (
  LOOKUP_TABLE_ENTRY_ID  OID DEFAULT nextval('LOOKUP_TABLE_ENTRY_SEQ')
                         PRIMARY KEY,
  LOOKUP_TABLE_ID        OID NOT NULL REFERENCES LOOKUP_TABLES
                         DEFERRABLE INITIALLY DEFERRED,
  VALUE                  VARCHAR(256) NOT NULL,
  LABEL                  VARCHAR(256)
);
CREATE INDEX LOOKUP_TABLE_ENTRIES_LT_ID ON LOOKUP_TABLE_ENTRIES(LOOKUP_TABLE_ID);
COMMENT ON TABLE LOOKUP_TABLE_ENTRIES IS
'This table and LOOKUP_TABLES make up a mechanism to restrict certain \
module inputs to a list of values.  This table identifies the values \
in a list, and LOOKUP_TABLES identifies the various value lists.';


CREATE SEQUENCE DATA_TABLE_SEQ;
CREATE TABLE DATA_TABLES (
  DATA_TABLE_ID  OID DEFAULT NEXTVAL('DATA_TABLE_SEQ') PRIMARY KEY,
  GRANULARITY    CHAR(1) NOT NULL
                 CHECK (GRANULARITY IN ('G','D','I','F')),
  TABLE_NAME     VARCHAR(64) NOT NULL,
  DESCRIPTION    TEXT
);
CREATE INDEX DATA_TABLES_NAME ON DATA_TABLES(TABLE_NAME);
COMMENT ON TABLE DATA_TABLES IS
'Each database table used to stored analysis results has an entry in \
the DATA_TABLES table.';
COMMENT ON COLUMN DATA_TABLES.GRANULARITY IS
'The kind of attribute table being refered to is represented by a \
single character - G, D, I or F corresponding to Global, Dataset, \
Image or Feature';


CREATE SEQUENCE DATA_COLUMN_SEQ;
CREATE TABLE DATA_COLUMNS (
  DATA_COLUMN_ID  OID DEFAULT NEXTVAL('DATA_COLUMN_SEQ') PRIMARY KEY,
  DATA_TABLE_ID   OID NOT NULL REFERENCES DATA_TABLES
                  DEFERRABLE INITIALLY DEFERRED,
  COLUMN_NAME     VARCHAR(64) NOT NULL,
  DESCRIPTION     TEXT,
  SQL_TYPE        VARCHAR(64) NOT NULL,
  REFERENCE_TYPE  VARCHAR(64),
  UNIQUE (DATA_TABLE_ID, COLUMN_NAME)
);
CREATE INDEX DATA_COLUMNS_TABLE ON DATA_COLUMNS(DATA_TABLE_ID);
CREATE INDEX DATA_COLUMNS_NAME ON DATA_COLUMNS(COLUMN_NAME);
COMMENT ON TABLE DATA_COLUMNS IS
'This table stores the names of the columns in each data table.  Each \
data table is required to declare three implicit columns, which are \
_not_ entered into the DATA_COLUMNS table.  They are ATTRIBUTE_ID; \
ANALYIS_ID; and one of DATASET_ID, IMAGE_ID, and FEATURE_ID, depending \
on the table\'s granularity.';


CREATE SEQUENCE SEMANTIC_TYPE_SEQ;
CREATE TABLE SEMANTIC_TYPES (
  SEMANTIC_TYPE_ID  OID DEFAULT nextval('SEMANTIC_TYPE_SEQ') PRIMARY KEY,
  NAME               VARCHAR(64) NOT NULL UNIQUE,
  GRANULARITY        CHAR(1) NOT NULL
                     CHECK (GRANULARITY IN ('G','D','I','F')),
  DESCRIPTION        TEXT
);
COMMENT ON TABLE SEMANTIC_TYPES IS
'This table stores the semantic types known to OME.  A semantic type \
is composed of one or more columns in a given table.  Future versions \
may allow attribute types to span tables.  Entries in \
SEMANTIC_ELEMENTS that share an SEMANTIC_TYPE_ID constitute a \
semantic type.  There is one row in this table for each semantic type \
known to OME.';
COMMENT ON COLUMN SEMANTIC_TYPES.GRANULARITY IS
'This column describes the granularity of the semantic type, wether it \
applies Globally or to a Dataset, Image or Feature.  This column must \
contain a single character, "G", "D", "I", or "F", and it must match \
the granularity specified in the DATA_TABLES that this types \
SEMANTIC_ELEMENTS refer to.';

CREATE SEQUENCE SEMANTIC_ELEMENT_SEQ;
CREATE TABLE SEMANTIC_ELEMENTS (
  SEMANTIC_ELEMENT_ID  OID DEFAULT NEXTVAL('SEMANTIC_ELEMENT_SEQ')
                       PRIMARY KEY,
  SEMANTIC_TYPE_ID    OID NOT NULL REFERENCES SEMANTIC_TYPES
                       DEFERRABLE INITIALLY DEFERRED,
  NAME                 VARCHAR(64) NOT NULL,
  DATA_COLUMN_ID       OID NOT NULL REFERENCES DATA_COLUMNS
                       DEFERRABLE INITIALLY DEFERRED,
  DESCRIPTION          TEXT,
  UNIQUE (SEMANTIC_TYPE_ID, NAME)
);
CREATE INDEX SEMANTIC_ELEMENTS_TYPE ON SEMANTIC_ELEMENTS(SEMANTIC_TYPE_ID);
CREATE INDEX SEMANTIC_ELEMENTS_COLUMN ON SEMANTIC_ELEMENTS(DATA_COLUMN_ID);
COMMENT ON TABLE SEMANTIC_ELEMENTS IS
'Attribute columns are the atomic types which constitute semantic \
types in OME.  A semantic type shares a common SEMANTIC_TYPE_ID which \
refers to a row in the SEMANTIC_TYPES tables.  A semantic type has \
one row in this table for each column (atomic type) it consists of.  \
The atomic types are specified by a DATA_COLUMN_ID refering to the \
DATA_COLUMNS table';

CREATE SEQUENCE ATTRIBUTE_SEQ;
