-- 000-OMECore.sql
-- ---------------------------------------
-- Author:  Ilya G. Goldberg (igg@nih.gov)
-- Copyright 1999-2002 Open Microscopy Environment (OME)
-- This file is part of OME.
-- ---------------------------------------
-- 
--     OME is free software; you can redistribute it and/or modify
--     it under the terms of the GNU General Public License as published by
--     the Free Software Foundation; either version 2 of the License, or
--     (at your option) any later version.
-- 
--     OME is distributed in the hope that it will be useful,
--     but WITHOUT ANY WARRANTY; without even the implied warranty of
--     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--     GNU General Public License for more details.
-- 
--     You should have received a copy of the GNU General Public License
--     along with OME; if not, write to the Free Software
--     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-- 


-- The tables listed in this file make up the core components of OME
-- The tables in the companion files are specific to analysis functions
-- of OME, and the outputs of these analyses.
-- 
-- Note that the first set of tables is quite primite right now.  The
-- layout is mainly to contain some basic information and to specify
-- the relationships.  In the future, many more columns will need to
-- be added to these tables, and possibly additional tables I haven't
-- thought of.
-- 
-- Some data type conventions:
--
-- Coordinates: Dataset coordinates are expressed as 2-byte integers,
-- or 4-byte floating points in the case of centroids or other
-- quantities that are the result of averages or other calculations.
-- Wavelegths are 2-byte integers and expressed in nm (maximum
-- wavelegth is 32767 nm or 32 um) The signal intensity is assumed to
-- be no greater than 2 billion counts (an INT4).  Integrals and other
-- intensity sums are 8-byte integers (INT8).  Someday someone will
-- make a camera that has more than 32767 x 32767 pixels, and collects
-- data to better than 31 bits.  That's the day I'll regret this
-- decision.  Also, right now there is no support for floating-point
-- pixel values.  In a way, floating point is way over-rated.  There
-- is more than enough dynamic range in a 4-byte integer to express
-- any raw or derived data (more than 9 significant figures) - in the
-- vast majority of cases, there is more than enough in a 1-byte
-- integer as long as the scale is properly selected.  To express
-- floating-point pixels, one could simply multiply (or divide) them
-- by an appropriate constant.
--
-- The datatypes are as defined in PostgreSQL convention.  Someday
-- I'll find a full listing of Oracle datatypes and map them
-- accordingly.  Also, I always try to specify the sizes of the
-- numeric datatypes (INT4 instead of INTEGER) to avoid any confusion,
-- and make implicit numerical limits explicit.
--
-- An additional point.  Primary keys are generated as sequences
-- instead of OIDs.  This makes them more visible to the user at the
-- expense of storage space.  Every row of every table has an OID,
-- which could easily serve as the primary key.  OIDs are not
-- generally displayed as part of column listings unless specifically
-- asked for.  The other disadvantage is that they are non-sequential
-- in a given table.  These disadvantages seem minor, but for now at
-- least, primary keys are generated from sequences specific for a
-- given table.  This may change.


-- CREATE SEQUENCE INSTRUMENT_SEQ;
-- CREATE TABLE INSTRUMENTS (
--   -- The primary key is a sequence which is automatically entered by default.
--   INSTRUMENT_ID      OID DEFAULT nextval ('INSTRUMENT_SEQ')
--                      PRIMARY KEY,
--   NAME               VARCHAR(30),
--   DESCRIPTION        TEXT
-- );
-- COMMENT ON TABLE INSTRUMENTS IS
-- 'The instruments table describes the microscope and mainly acts as a \
-- nexus for the many-to-one relationships of the components that \
-- constitute it - i.e. Objectives, Filters, etc.';

CREATE SEQUENCE EXPERIMENTER_SEQ;
CREATE TABLE EXPERIMENTERS (
  ATTRIBUTE_ID   OID DEFAULT NEXTVAL('ATTRIBUTE_SEQ') PRIMARY KEY,
  MODULE_EXECUTION_ID    OID,
   OME_NAME           VARCHAR(30),
   FIRSTNAME          VARCHAR(30),
   LASTNAME           VARCHAR(30),
   EMAIL              VARCHAR(50),
   password           varchar(64),
   GROUP_ID           OID,  -- This refers to the table groups, which is not 
                            -- yet defined.  Groups in turn has references
                            -- to experimenters, creating a circular reference.  
                            -- We create the foreign key constraint after
                            -- the experimenters table.
   DATA_DIR           VARCHAR(256),
   INSTITUTION              VARCHAR(256),
   CONSTRAINT unique_ome_name UNIQUE(OME_NAME)
);
COMMENT ON TABLE EXPERIMENTERS IS
'The experimenters table defines the people who perform imaging \
experiments.  Each experimenter may belong to one Group';



-- These are research groups that use OME.  By default, the Image's
-- group is set to the Experimenter's group at the time of import.
-- The group leader and contact refer to the experimenters table.

CREATE SEQUENCE GROUP_SEQ;
CREATE TABLE GROUPS (
  ATTRIBUTE_ID    OID    DEFAULT nextval ('ATTRIBUTE_SEQ') PRIMARY KEY,
  MODULE_EXECUTION_ID     OID,
  NAME        VARCHAR(30),
  LEADER      OID REFERENCES EXPERIMENTERS
              DEFERRABLE INITIALLY DEFERRED,
  CONTACT     OID REFERENCES EXPERIMENTERS
              DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE GROUPS IS
'The groups table defines groups of experimenters.  This is can be a \
lab or project group.  It is not meant to represent an institution or \
a company';

-- ALTER TABLE experimenters ADD CONSTRAINT groupfk
--   FOREIGN KEY (GROUP_ID) REFERENCES GROUPS(GROUP_ID)
--   DEFERRABLE INITIALLY DEFERRED;


--
-- IGG added 9/12/02
-- A dataset is an arbitrarily named collection of images.  Datasets
-- have experimenter and group ownership.  The experimenter and group
-- are set to the experimenter and experimenter's group at the time of
-- Dataset creation.  They don't have to match the image's
-- experimenter and group.  A dataset represents a collection of
-- images that are always analyzed as a unit.  Images can belong to
-- more than one dataset.  Datasets in turn can belong to more than
-- one project.

CREATE SEQUENCE DATASET_SEQ;
CREATE TABLE DATASETS (
  DATASET_ID     OID DEFAULT nextval('DATASET_SEQ') PRIMARY KEY,
  NAME           VARCHAR(256) NOT NULL,
  OWNER_ID       OID NOT NULL REFERENCES EXPERIMENTERS
                 DEFERRABLE INITIALLY DEFERRED,
  GROUP_ID       OID REFERENCES GROUPS
                 DEFERRABLE INITIALLY DEFERRED,
  DESCRIPTION    TEXT,
  LOCKED         BOOLEAN NOT NULL DEFAULT FALSE
);
COMMENT ON TABLE DATASETS IS
'Datasets are groups of images representing a "computational unit".  \
Once an analysis is performed on a Dataset, the LOCKED flag is set and \
the list of images that compose a dataset may no longer be changed.  \
Images may belong to one or more Datasets.  This many-to-many \
relationship is specified by the IMAGE_DATASET_MAP table';


--
-- IGG added 9/12/02
-- A project is an arbitrarily named collection of datasets.  Projects
-- have experimenter and group ownership.

CREATE SEQUENCE PROJECT_SEQ;
CREATE TABLE PROJECTS (
  PROJECT_ID         OID   DEFAULT nextval ('PROJECT_SEQ') PRIMARY KEY,
  NAME               VARCHAR(64) NOT NULL,
  OWNER_ID           OID NOT NULL REFERENCES EXPERIMENTERS
                     DEFERRABLE INITIALLY DEFERRED,
  GROUP_ID           OID REFERENCES GROUPS
                     DEFERRABLE INITIALLY DEFERRED,
  VIEW               VARCHAR(64),  -- The name of the most recently
                                   -- used view for this project.
  DESCRIPTION        TEXT
);
COMMENT ON TABLE PROJECTS IS
'A Project is simply a group of Datasets to be used as an \
organizational aid to the user.  Datasets may belong to one or more \
Projects.  This many-to-many relationship is specified by the \
PROJECT_DATASET_MAP table.';


--
-- IGG added 9/12/02
-- Projects and datasets have a many to many relationship represented
-- by this map.

CREATE TABLE PROJECT_DATASET_MAP (
  PROJECT_ID   OID NOT NULL REFERENCES PROJECTS
               DEFERRABLE INITIALLY DEFERRED,
  DATASET_ID   OID NOT NULL REFERENCES DATASETS
               DEFERRABLE INITIALLY DEFERRED,
  PRIMARY KEY (PROJECT_ID, DATASET_ID)
);
COMMENT ON TABLE PROJECT_DATASET_MAP IS
'This table specifies the many-to-many relationship between Datasets \
and Projects.';


--
-- These are display settings for images.  Each image points to a
-- default setting.  Also, each user session has a default setting.
-- By default, the image setting over-rides the session setting.
-- There is also a system setting which is over-ridden by any session
-- or image settings.  This is mean to control the viewer, but also
-- serves as a mechanism for generating a thumbnail in a
-- non-interactive fashion.  The viewer is supposed to allow the user
-- to set his session setting, the image's setting, etc.

 CREATE SEQUENCE IMAGE_DISPLAY_SETTINGS_SEQ;
 CREATE TABLE IMAGE_DISPLAY_SETTINGS (
   IMAGE_DISPLAY_SETTINGS_ID  OID DEFAULT nextval ('IMAGE_DISPLAY_SETTINGS_SEQ') 
                              PRIMARY KEY,
   URL                        TEXT  -- This is meant to be a GET request to a
                                    -- cgi, but can be static also.
 );
 COMMENT ON TABLE IMAGE_DISPLAY_SETTINGS IS
 'This table will eventually contain various "hints" to a graphical \
 image viewer.';


-- --
-- -- IGG added 9/12/02
-- -- This maintains a path to a repository-format file - a raw pixel dump.

CREATE SEQUENCE REPOSITORY_SEQ;
CREATE TABLE REPOSITORIES (
  ATTRIBUTE_ID   OID DEFAULT nextval('ATTRIBUTE_SEQ') PRIMARY KEY,
  MODULE_EXECUTION_ID    OID,
  PATH            VARCHAR(256) NOT NULL UNIQUE
);
COMMENT ON TABLE REPOSITORIES IS
'OME Image pixels are stored in repository format - a 5-D pixel dump \
in XYZWT order.  The Repositories table maintains directory paths to \
one or more repositories';


--
-- This is the main image object, which may be composed of several
-- files.  A full image description make use of several auxilliary
-- tables listed below.

CREATE SEQUENCE IMAGE_SEQ;
CREATE TABLE IMAGES (
  IMAGE_ID          OID DEFAULT nextval('IMAGE_SEQ') PRIMARY KEY,
  IMAGE_GUID        VARCHAR(256),     --  See the OME Image GUID Spec
  NAME              VARCHAR(256) NOT NULL,
  --FILE_SHA1         CHAR (40),        -- SHA1 digest of the file.
  --DISPLAY_SETTINGS  OID REFERENCES image_display_settings
  --                  DEFERRABLE INITIALLY DEFERRED,
  DESCRIPTION       TEXT,        
  --INSTRUMENT_ID     OID REFERENCES INSTRUMENTS
  --                  DEFERRABLE INITIALLY DEFERRED,
  --LENS_ID           VARCHAR(30), -- ID of lens that took input images
  EXPERIMENTER_ID   OID NOT NULL REFERENCES EXPERIMENTERS
                    DEFERRABLE INITIALLY DEFERRED,
  GROUP_ID          OID REFERENCES GROUPS,
  CREATED           TIMESTAMP NOT NULL,
  INSERTED          TIMESTAMP NOT NULL
  --IMAGE_TYPE        OID,    -- Circular reference, the fkey will be 
  --                          -- declared after the IMAGE_TYPES table
  --REPOSITORY_ID     OID NOT NULL REFERENCES REPOSITORIES
  --                  DEFERRABLE INITIALLY DEFERRED,
  --PATH              VARCHAR(256) NOT NULL
);
CREATE INDEX images_image_id_idx ON images (image_id);
CREATE INDEX images_name_idx ON images (name);
COMMENT ON TABLE IMAGES IS
'The Images table contains information common to all images, including \
both the local ID and the globally unique ID, and several references \
to other tables with additional image information.';


--
-- IGG added 9/12/02
-- This table is a map between images and datasets.  Images may belong
-- to more than one dataset.

CREATE TABLE IMAGE_DATASET_MAP (
  IMAGE_ID    OID NOT NULL REFERENCES IMAGES
              DEFERRABLE INITIALLY DEFERRED,
  DATASET_ID  OID NOT NULL REFERENCES DATASETS
              DEFERRABLE INITIALLY DEFERRED,
  PRIMARY KEY (IMAGE_ID, DATASET_ID)
);
CREATE INDEX images_datasets_image_idx ON IMAGE_DATASET_MAP (IMAGE_ID);
CREATE INDEX images_datasets_datasets_idx ON IMAGE_DATASET_MAP (DATASET_ID);
COMMENT ON TABLE IMAGE_DATASET_MAP IS
'This table represents the many-to-many relationship between Images \
and Datasets';


--
--  Datasets are not necessarily contained in a single file.  Files do
--  contain contiguous data.  For multiple files, how do we put them
--  together into images?  Do the files need to know where in the
--  image they belong, or is it up to the images to sort out their
--  files?  An OO model might favor that the specific image
--  implementation should worry about its specific file layout.  The
--  OME DB schema is supposed to be a fully relational data model, so
--  the file 'object' should know how it fits into the image.  This
--  has the additional benefit that now we've standardized file->image
--  mappings in a generic way.  At least for 5-D XYZWT images.
--  Additionally, this is no more painful or messy of an OO model.
--  The assumption with this mapping is that files contain contiguous
--  data, but the individual contiguous blocks can be distributed
--  among an arbitrary set of files, including the pathological
--  condition of one pixel per file (!).

CREATE TABLE IMAGE_FILES_XYZWT (
  IMAGE_ID        OID REFERENCES IMAGES
                  DEFERRABLE INITIALLY DEFERRED,
  FILE_SHA1       CHAR (40), -- SHA1 digest of the file.
  BIGENDIAN       BOOLEAN,   -- whether the file is big- or little endian
  PATH            VARCHAR(256),
  HOST            VARCHAR(256),
  URL             VARCHAR(256),
  X_START         SMALLINT,      --  XYZ are pixel coordinates
  X_STOP          SMALLINT,
  Y_START         SMALLINT,
  Y_STOP          SMALLINT,
  Z_START         SMALLINT,
  Z_STOP          SMALLINT,
  W_START         SMALLINT,      --  W is wave number
  W_STOP          SMALLINT,
  T_START         SMALLINT,      --  T is time point
  T_STOP          SMALLINT
);
COMMENT ON TABLE IMAGE_FILES_XYZWT IS
'Entries in this table store locations of the non-OME image files used \
to constitute an OME image.  Using this table, it is possible to \
assemble an aritrary number of images into a single 5-D OME Image';



--
-- This table specifies the image types that OME knows about and how
-- they are implemented.

CREATE SEQUENCE IMAGE_TYPE_SEQ;
CREATE TABLE IMAGE_TYPES (
  IMAGE_TYPE           OID DEFAULT nextval ('IMAGE_TYPE_SEQ') PRIMARY KEY,
  TYPE_NAME            VARCHAR(64),
  IMPLEMENTATION_LANG  VARCHAR(32),   -- How the interface is implemented
                                      -- i.e. 'Perl', 'Java', 'C', 'C++', etc.
  CLASSPATH            VARCHAR(255),  -- in the implementation language 
                                      -- ('OMEDataset::XYZWT',
                                      -- 'org.ome.image.xyzwt', etc.)
  ATTRIBUTES_TABLE     VARCHAR(64)
);
--
--ALTER TABLE IMAGES ADD CONSTRAINT image_typefk FOREIGN KEY (IMAGE_TYPE)
--  REFERENCES IMAGE_TYPES(IMAGE_TYPE)
--  DEFERRABLE INITIALLY DEFERRED;
COMMENT ON TABLE IMAGE_TYPES IS
'This table allows OME to keep track of the formats of the non-OME \
image files, and the various programs used to import them into OME';


--
-- This table maintains a list of expired files - files to be deleted
-- from the file system as one of the periodic database maintenance
-- chores.  Entries in this table should be deleted after the files
-- are actually removed from the system.  Attributes that maintain
-- data in files on the filesystem should create rules such as:
-- CREATE RULE expire_mytable_files AS
--   ON DELETE TO mytable DO
--   INSERT INTO expired_files VALUES (OLD.path || OLD.name);
CREATE TABLE EXPIRED_FILES (
  fullpath    varchar(255)
);
COMMENT ON TABLE EXPIRED_FILES IS
'This table keeps track of any auxiliary file used by OME that is no \
longer needed, but is not yet deleted from the file system.';


-- This is a table with project and image view templates.  variable
-- substitution will occur wherever template matches $(ProjectID) or
-- $(DatasetID) The type is either 'Dataset' or 'Project'.  Possibly
-- other types to follow.
CREATE TABLE VIEW_TEMPLATES (
  NAME          VARCHAR(32),
  DESCRIPTION   TEXT,
  TEMPLATE      TEXT,
  TYPE          VARCHAR(32)
  --OWNER_ID      OID REFERENCES EXPERIMENTERS
  --              DEFERRABLE INITIALLY DEFERRED,
  --GROUP_ID      OID REFERENCES GROUPS
  --              DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE VIEW_TEMPLATES IS
'This table is used to store SQL templates for generating views.  The \
SQL is stored in the template column and may include variables in the \
form $(VariableName) that will be substituted by the apropriate value \
when the view is created.  Commonly used variables include \
$(ExperimenterID), $(ImageID), $(ProjectID), $(DatasetID), etc.';


-- Make a rule or trigger to delete old entries where ome_name is the
-- same as newly added ome_name, i.e. one session per user.  This is
-- currently done in OMElogin.pl OMElogin.pl preserves a user's
-- settings from a previous session.

CREATE SEQUENCE SESSION_SEQ; 
CREATE TABLE OME_SESSIONS (
  SESSION_ID         OID DEFAULT nextval ('SESSION_SEQ') PRIMARY KEY,
  SESSION_KEY        VARCHAR(36),  -- The client's state info (not the
                                   -- user's state info, which is SESSION_ID)
  EXPERIMENTER_ID    OID
                     NOT NULL REFERENCES EXPERIMENTERS
                     DEFERRABLE INITIALLY DEFERRED,
  HOST               VARCHAR(256),                 -- The remote host
  PROJECT_ID         OID REFERENCES PROJECTS
                     DEFERRABLE INITIALLY DEFERRED,
  DATASET_ID         OID REFERENCES DATASETS
                     DEFERRABLE INITIALLY DEFERRED,
  MODULE_EXECUTION_ID        OID,
  IMAGE_VIEW         TEXT,  -- A select clause ending with where
  FEATURE_VIEW       TEXT,  -- A select clause ending with where - the view
                            -- name will be the imageName.
  MODULE_EXECUTION           VARCHAR(64),  -- A relative URL to the page in the 
                                   -- MainFrame - should be a program ID?
  DISPLAY_SETTINGS   OID REFERENCES IMAGE_DISPLAY_SETTINGS
                     DEFERRABLE INITIALLY DEFERRED,
  LAST_ACCESS        TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
  STARTED            TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);
COMMENT ON TABLE OME_SESSIONS IS
'This table is used to store various information about the user\'s \
session.  This session never expires, though the session_key may have \
a very short lifetime (minutes).  When a new session_key is generated \
for a session, the appropriate row is updated with the new value.';

CREATE TABLE OME_SESSIONS_IMAGES (
  IMAGE_ID    OID REFERENCES IMAGES
              DEFERRABLE INITIALLY DEFERRED,
  SESSION_ID  OID REFERENCES OME_SESSIONS
              DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE OME_SESSIONS_IMAGES IS
'Multiple Images, Projects, and/or Datasets may be associated with a \
session.  This table lists the Images associated with a particular \
Session.';

CREATE TABLE CONFIGURATION (
  CONFIGURATION_ID   SMALLINT DEFAULT 1  -- Always set to 1 - should only
                     PRIMARY KEY,        -- be 1 entry in this table
  MAC_ADDRESS        VARCHAR(20),        -- The original host's ethernet
                                         --  card's MAC address
  DB_INSTANCE        CHAR(6),            -- base64-encoded, network-order result of Perl's time().
  LSID_AUTHORITY     VARCHAR(256),       -- The authority part of local LSIDs (i.e. mycomputer.mydomain.com)
  TMP_DIR            VARCHAR(256),       -- Scratch space
  XML_DIR            VARCHAR(256),       -- Location for xml files
  BIN_DIR            VARCHAR(256),       -- Location for OME binaries
  REPOSITORY_DIR     VARCHAR(256),       -- root of image repository
  OME_ROOT           VARCHAR(256),       -- root of OME executables,
  IMPORT_MODULE      OID,                -- import module
  IMPORT_CHAIN       OID,                -- import chain
  DISPLAY_SETTINGS   OID REFERENCES IMAGE_DISPLAY_SETTINGS
                     DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE CONFIGURATION IS
'This table holds this OME instance\'s configuration parameters. It \
will normally be filled in by script or installer queries at OME \
install time.';


-- added by Josiah March 21, 2003
CREATE TABLE VIEWER_PREFERENCES (
  ATTRIBUTE_ID      OID DEFAULT NEXTVAL ('ATTRIBUTE_SEQ') PRIMARY KEY,
  EXPERIMENTER_ID   OID NOT NULL
                    REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED, 
  TOOLBOX_SCALE     REAL
);
COMMENT ON TABLE VIEWER_PREFERENCES IS
'This table stores viewer preferences for an experimenter.';

