-- 010-Analysis.sql
-- -------------------------------------------------
-- Authors:  Douglas Creager (dcreager@alum.mit.edu)
--           Ilya G. Goldberg (igg@nih.gov)
-- Copyright 1999-2002 Open Microscopy Environment (OME)
-- This file is part of OME.
-- -------------------------------------------------
-- 
--     OME is free software; you can redistribute it and/or modify
--     it under the terms of the GNU General Public License as published by
--     the Free Software Foundation; either version 2 of the License, or
--     (at your option) any later version.
-- 
--     OME is distributed in the hope that it will be useful,
--     but WITHOUT ANY WARRANTY; without even the implied warranty of
--     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--     GNU General Public License for more details.
-- 
--     You should have received a copy of the GNU General Public License
--     along with OME; if not, write to the Free Software
--     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-- 


--
-- *****
-- 17 May 2002 (DC)
--   - Changed PROGRAMS table to require analyses to implement a Perl
--     class (specified by the LOCATION field).
--   - Eventually, this will migrate to an
--     XML-RPC/SOAP/etc. implementation, and the LOCATION field will
--     be interpreted differently (as a URL, most likely)
--
-- 24 September 2002
--   - Removed INPUT_TABLE column since this is now specified in
--     the formal inputs table

CREATE SEQUENCE MODULE_CATEGORY_SEQ;
CREATE TABLE MODULE_CATEGORIES (
       CATEGORY_ID         OID DEFAULT NEXTVAL('MODULE_CATEGORY_SEQ')
                           PRIMARY KEY,
       NAME                VARCHAR(64) NOT NULL,
       PARENT_CATEGORY_ID  OID REFERENCES MODULE_CATEGORIES
                           DEFERRABLE INITIALLY DEFERRED,
       DESCRIPTION         TEXT,
       UNIQUE (PARENT_CATEGORY_ID, NAME)
);
CREATE INDEX PROG_CATEGORIES_NAME ON MODULE_CATEGORIES(PARENT_CATEGORY_ID);

CREATE SEQUENCE MODULE_SEQ;
CREATE TABLE MODULES (
       MODULE_ID             OID DEFAULT NEXTVAL('MODULE_SEQ') PRIMARY KEY,
       NAME           VARCHAR(64) NOT NULL,
       DESCRIPTION            TEXT,
       LOCATION               VARCHAR(128) NOT NULL,
       MODULE_TYPE            VARCHAR(128) NOT NULL,
       CATEGORY               OID REFERENCES MODULE_CATEGORIES
                              DEFERRABLE INITIALLY DEFERRED,
       DEFAULT_ITERATOR       VARCHAR(128),
       NEW_FEATURE_TAG        VARCHAR(128),
       EXECUTION_INSTRUCTIONS TEXT
);
CREATE INDEX MODULES_NAME ON MODULES(NAME);
CREATE INDEX MODULES_CATEGORY ON MODULES(CATEGORY);
CREATE INDEX MODULES_LOCATION ON MODULES(LOCATION);
CREATE INDEX MODULES_MODULE_TYPE ON MODULES(MODULE_TYPE);
COMMENT ON TABLE MODULES IS
'This table specifies an analysis module registered with OME.  There \
is only one entry in this table for each module known to OME, unlike \
the MODULE_EXECUTIONS table, which has one entry per module run.';
COMMENT ON COLUMN MODULES.CATEGORY IS
'Groups modules by category to organize things for the user.';
COMMENT ON COLUMN MODULES.MODULE_TYPE IS
'Specifies the analysis handler used to execute this module.  Separate \
analysis handlers are written for each "class" of module, e.g., \
interal Perl script, external command-line module, external SOAP \
service.  This field should refer to a Perl class, visible to OME, \
implementing the OME::Analysis::Handler interface.';
COMMENT ON COLUMN MODULES.LOCATION IS
'The physical location of the module.  This has different meanings for \
different analysis handlers.  For instance, the \
OME::Analysis::CLIHandler module type expects this field to be the \
filename of a command-line executable.  OME::Analysis::PerlHandler \
expects this to be the name of a Perl class.';
COMMENT ON COLUMN MODULES.EXECUTION_INSTRUCTIONS IS
'An XML fragment from the XML OME Module specification that describes \
how to execute the module.  It is used to pass handler-specific \
parameters to the module\'s analysis handler.';
COMMENT ON COLUMN MODULES.DEFAULT_ITERATOR IS
'If this module takes in feature-granular inputs, this column \
specifies which features in the feature hierarchy are presented as \
inputs to the module.  It can be overridden at chain-design-time via \
the ANALYSIS_CHAIN_NODES.ITERATOR_TAG column.  For more information, \
see the FEATURES table.';
COMMENT ON COLUMN MODULES.NEW_FEATURE_TAG IS
'If new features are created by this module, they will be given a tag \
specified by this column.  They will be children of the features \
specified by DEFAULT_ITERATOR.  For more information, see the FEATURES \
table.';

ALTER TABLE CONFIGURATION ADD CONSTRAINT IMPORT_MODULEFK
  FOREIGN KEY (IMPORT_MODULE) REFERENCES MODULES
  DEFERRABLE INITIALLY DEFERRED;

-- 24 September 2002 (DC)
-- New tables to store the defined inputs/outputs of the analysis modules.

CREATE SEQUENCE FORMAL_INPUT_SEQ;
CREATE TABLE FORMAL_INPUTS (
  FORMAL_INPUT_ID    OID DEFAULT nextval('FORMAL_INPUT_SEQ')
                     PRIMARY KEY,
  MODULE_ID         OID NOT NULL REFERENCES MODULES
                     DEFERRABLE INITIALLY DEFERRED,
  NAME               VARCHAR(64) NOT NULL,
  DESCRIPTION        TEXT,
  OPTIONAL           BOOLEAN DEFAULT FALSE,
  LIST               BOOLEAN DEFAULT TRUE,
  SEMANTIC_TYPE_ID  OID NOT NULL REFERENCES SEMANTIC_TYPES
                     DEFERRABLE INITIALLY DEFERRED,
  LOOKUP_TABLE_ID    OID REFERENCES LOOKUP_TABLES
                     DEFERRABLE INITIALLY DEFERRED,
  USER_DEFINED       BOOLEAN DEFAULT FALSE,
  UNIQUE (MODULE_ID,NAME)
);
CREATE INDEX FORMAL_INPUTS_MODULE_ID ON FORMAL_INPUTS(MODULE_ID);
CREATE INDEX FORMAL_INPUTS_NAME ON FORMAL_INPUTS(NAME);
CREATE INDEX FORMAL_INPUTS_TYPE ON FORMAL_INPUTS(SEMANTIC_TYPE_ID);
COMMENT ON TABLE FORMAL_INPUTS IS
'The FORMAL_INPUTS table declares the semantic types used as inputs by \
an analysis module.  Each input maps to exactly one entry in the \
SEMANTIC_TYPES table.  A table of possible user input values may be \
specified by referring to an entry in LOOKUP_TABLES.  FORMAL_INPUTS \
has a many-to-one relationship with the MODULES table:  A MODULE has \
zero or more FORMAL_INPUTS';


CREATE SEQUENCE FORMAL_OUTPUT_SEQ;
CREATE TABLE FORMAL_OUTPUTS (
  FORMAL_OUTPUT_ID   OID DEFAULT nextval('FORMAL_OUTPUT_SEQ') PRIMARY KEY,
  MODULE_ID         OID NOT NULL REFERENCES MODULES
                     DEFERRABLE INITIALLY DEFERRED,
  NAME               VARCHAR(64) NOT NULL,
  DESCRIPTION        TEXT,
  OPTIONAL           BOOLEAN DEFAULT FALSE,
  LIST               BOOLEAN DEFAULT TRUE,
  SEMANTIC_TYPE_ID  OID REFERENCES SEMANTIC_TYPES
                     DEFERRABLE INITIALLY DEFERRED,
  FEATURE_TAG        VARCHAR(128),
  UNIQUE (MODULE_ID,NAME),
  UNIQUE (MODULE_ID,SEMANTIC_TYPE_ID)
);
CREATE INDEX FORMAL_OUTPUTS_MODULE_ID ON FORMAL_OUTPUTS(MODULE_ID);
CREATE INDEX FORMAL_OUTPUTS_NAME ON FORMAL_OUTPUTS(NAME);
CREATE INDEX FORMAL_OUTPUTS_TYPE ON FORMAL_OUTPUTS(SEMANTIC_TYPE_ID);
COMMENT ON TABLE FORMAL_OUTPUTS IS
'The FORMAL_OUTPUTS table declares the semantic types generated as \
outputs from a analysis modules.  Each output maps to exactly one \
entry in the SEMANTIC_TYPES table.  FORMAL_OUTPUTS has a many-to-one \
relationship with the MODULES table: A MODULE has one or more \
FORMAL_OUTPUTS';


-- analysis views (a particular graph of modules linked via inputs/outputs)

CREATE SEQUENCE ANALYSIS_CHAIN_SEQ;
CREATE TABLE ANALYSIS_CHAINS (
  ANALYSIS_CHAIN_ID  OID DEFAULT nextval('ANALYSIS_CHAIN_SEQ') PRIMARY KEY,
  OWNER             OID NOT NULL
                    REFERENCES EXPERIMENTERS
                    DEFERRABLE INITIALLY DEFERRED,
  NAME              VARCHAR(64) NOT NULL,
  DESCRIPTION       TEXT,
  LOCKED            BOOLEAN NOT NULL DEFAULT FALSE
);
CREATE INDEX ANALYSIS_CHAINS_OWNER ON ANALYSIS_CHAINS(OWNER);
COMMENT ON TABLE ANALYSIS_CHAINS IS
'An "analysis chain" (or equivalently, "analysis view") is a directed \
graph of analysis modules liked via their formal inputs/outputs.  It \
represents an execution plan for a chain of modules.  Multiple \
instances of a given analysis module can exist in a single chain.';

ALTER TABLE CONFIGURATION ADD CONSTRAINT IMPORT_CHAINFK
  FOREIGN KEY (IMPORT_CHAIN) REFERENCES ANALYSIS_CHAINS
  DEFERRABLE INITIALLY DEFERRED;

CREATE SEQUENCE ANALYSIS_CHAIN_NODES_SEQ;
CREATE TABLE ANALYSIS_CHAIN_NODES (
  ANALYSIS_CHAIN_NODE_ID  OID DEFAULT nextval('ANALYSIS_CHAIN_NODES_SEQ')
                         PRIMARY KEY, 
  ANALYSIS_CHAIN_ID       OID NOT NULL REFERENCES ANALYSIS_CHAINS
                         DEFERRABLE INITIALLY DEFERRED,
  MODULE_ID             OID NOT NULL REFERENCES MODULES
                         DEFERRABLE INITIALLY DEFERRED,
  ITERATOR_TAG           VARCHAR(128),
  NEW_FEATURE_TAG        VARCHAR(128)
);
CREATE INDEX ANALYSIS_CHAIN_NODES_VIEW ON ANALYSIS_CHAIN_NODES(ANALYSIS_CHAIN_ID);
CREATE INDEX ANALYSIS_CHAIN_NODES_MODULE ON ANALYSIS_CHAIN_NODES(MODULE_ID);
COMMENT ON TABLE ANALYSIS_CHAIN_NODES IS
'A single node within an analysis chain.  Multiple nodes in a single \
chain can refer to the same MODULE.';
COMMENT ON COLUMN ANALYSIS_CHAIN_NODES.MODULE_ID IS
'The analysis module which is run at this stage of the analysis \
chain.';
COMMENT ON COLUMN ANALYSIS_CHAIN_NODES.ITERATOR_TAG IS
'Overrides the value defined in MODULES.DEFAULT_ITERATOR_TAG.  See \
the FEATURES table for more information.';
COMMENT ON COLUMN ANALYSIS_CHAIN_NODES.NEW_FEATURE_TAG IS
'Overrides the value defined in MODULES.NEW_FEATURE_TAG.  See the \
FEATURES table for more information.';

CREATE SEQUENCE ANALYSIS_CHAIN_LINKS_SEQ;
CREATE TABLE ANALYSIS_CHAIN_LINKS (
  ANALYSIS_CHAIN_LINK_ID  OID DEFAULT nextval('ANALYSIS_CHAIN_LINKS_SEQ')
                         PRIMARY KEY,
  ANALYSIS_CHAIN_ID       OID NOT NULL REFERENCES ANALYSIS_CHAINS
                         DEFERRABLE INITIALLY DEFERRED,
  FROM_NODE              OID NOT NULL REFERENCES ANALYSIS_CHAIN_NODES
                         DEFERRABLE INITIALLY DEFERRED,
  FROM_OUTPUT            OID NOT NULL REFERENCES FORMAL_OUTPUTS
                         DEFERRABLE INITIALLY DEFERRED,
  TO_NODE                OID NOT NULL REFERENCES ANALYSIS_CHAIN_NODES
                         DEFERRABLE INITIALLY DEFERRED,
  TO_INPUT               OID NOT NULL REFERENCES FORMAL_INPUTS
                         DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX ANALYSIS_CHAIN_LINKS_VIEW ON ANALYSIS_CHAIN_LINKS(ANALYSIS_CHAIN_ID);
CREATE INDEX ANALYSIS_CHAIN_LINKS_FROM_NODE ON ANALYSIS_CHAIN_LINKS(FROM_NODE);
CREATE INDEX ANALYSIS_CHAIN_LINKS_OUTPUT ON ANALYSIS_CHAIN_LINKS(FROM_OUTPUT);
CREATE INDEX ANALYSIS_CHAIN_LINKS_TO_NODE ON ANALYSIS_CHAIN_LINKS(TO_NODE);
CREATE INDEX ANALYSIS_CHAIN_LINKS_inPUT ON ANALYSIS_CHAIN_LINKS(TO_INPUT);
COMMENT ON TABLE ANALYSIS_CHAIN_LINKS IS
'Defines the edges in an analysis chain.  Links are only allowed \
between modules if they connect outputs and inputs of the same type:  \
i.e., the SEMANTIC_TYPE_ID\'s of the FROM_OUTPUT and TO_INPUT must \
match.  Further, FROM_NODE and TO_NODE must be members of the same \
ANALYSIS_CHAIN.';

CREATE SEQUENCE ANALYSIS_PATH_SEQ;
CREATE TABLE ANALYSIS_PATHS (
  PATH_ID           OID DEFAULT nextval('ANALYSIS_PATH_SEQ') PRIMARY KEY,
  PATH_LENGTH       INTEGER NOT NULL,
  ANALYSIS_CHAIN_ID  OID NOT NULL REFERENCES ANALYSIS_CHAINS
                    DEFERRABLE INITIALLY DEFERRED
);

CREATE TABLE ANALYSIS_PATH_MAP (
  PATH_ID               OID NOT NULL REFERENCES ANALYSIS_PATHS
                        DEFERRABLE INITIALLY DEFERRED,
  PATH_ORDER            INTEGER NOT NULL,
  ANALYSIS_CHAIN_NODE_ID OID NOT NULL REFERENCES ANALYSIS_CHAIN_NODES
                        DEFERRABLE INITIALLY DEFERRED,
  PRIMARY KEY (PATH_ID, PATH_ORDER)
);
CREATE INDEX ANALYSIS_PATH_PATH ON ANALYSIS_PATH_MAP(PATH_ID);
CREATE INDEX ANALYSIS_PATH_NODE ON ANALYSIS_PATH_MAP(ANALYSIS_CHAIN_NODE_ID);
COMMENT ON TABLE ANALYSIS_PATH_MAP IS
'This table links a path through a directed graph of analyses to \
specific analyses.  Paths traverse the directed graph of analyses from \
the root node to a specific leaf node.  There is exactly one unique \
path between each root node and each leaf node.  Due to branching and \
merging in this analysis graph, analyses may belong to more than one \
path.  A given path is composed of many analyses.  This table \
specifies the many-to-many relationship between paths and analyses.';


-- analyses - actual runs of a given module
CREATE SEQUENCE MODULE_EXECUTION_SEQ;
CREATE TABLE MODULE_EXECUTIONS (
  MODULE_EXECUTION_ID     OID DEFAULT nextval ('MODULE_EXECUTION_SEQ') PRIMARY KEY,
  MODULE_ID      OID REFERENCES MODULES
                  DEFERRABLE INITIALLY DEFERRED,
  DEPENDENCE      CHAR(1) NOT NULL
                  CHECK (DEPENDENCE IN ('G','D','I')),
  DATASET_ID      OID NOT NULL REFERENCES DATASETS
                  DEFERRABLE INITIALLY DEFERRED,
  TIMESTAMP       TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
  TOTAL_TIME      FLOAT,
  ATTRIBUTE_SORT_TIME    FLOAT,
  ATTRIBUTE_DB_TIME      FLOAT,
  ATTRIBUTE_CREATE_TIME  FLOAT,
  STATUS          VARCHAR(16)
);
CREATE INDEX MODULE_EXECUTIONS_MODULE_ID ON MODULE_EXECUTIONS(MODULE_ID);
--CREATE INDEX MODULE_EXECUTIONS_EXPERIMENTER_ID ON MODULE_EXECUTIONS(EXPERIMENTER_ID);
CREATE INDEX MODULE_EXECUTIONS_DATASET_ID ON MODULE_EXECUTIONS(DATASET_ID);
COMMENT ON TABLE MODULE_EXECUTIONS IS
'This table defines a "run" of a particular module against a \
particular dataset.';
COMMENT ON COLUMN MODULE_EXECUTIONS.DEPENDENCE IS
'This column specifies the highest level of the dataset-image-feature \
hierarchy that an analysis is dependent on.  DEPENDENCE can contain a \
single character, G, D or I corresponding to Global, Dataset and \
Image.';

ALTER TABLE EXPERIMENTERS ADD CONSTRAINT ANALYSISFK
  FOREIGN KEY (MODULE_EXECUTION_ID) REFERENCES MODULE_EXECUTIONS
  DEFERRABLE INITIALLY DEFERRED;


-- CREATE SEQUENCE ACTUAL_OUTPUT_SEQ;
-- CREATE TABLE ACTUAL_OUTPUTS (
--   ACTUAL_OUTPUT_ID  OID DEFAULT nextval('ACTUAL_OUTPUT_SEQ') PRIMARY KEY,
--   MODULE_EXECUTION_ID       OID NOT NULL REFERENCES MODULE_EXECUTIONS
--                     DEFERRABLE INITIALLY DEFERRED,
--   FORMAL_OUTPUT_ID  OID NOT NULL REFERENCES FORMAL_OUTPUTS
--                     DEFERRABLE INITIALLY DEFERRED
-- );
-- CREATE INDEX ACTUAL_OUTPUTS_MODULE_EXECUTION_ID ON ACTUAL_OUTPUTS(MODULE_EXECUTION_ID);
-- CREATE INDEX ACTUAL_OUTPUTS_FO_ID ON ACTUAL_OUTPUTS(FORMAL_OUTPUT_ID);
-- COMMENT ON TABLE ACTUAL_OUTPUTS IS
-- 'This table stores the actual outputs that were produced by a specific analysis by storing both the data type declaration (FORMAL_OUTPUT_ID).  Each data value generated links back to this table.';


CREATE SEQUENCE ACTUAL_INPUT_SEQ;
CREATE TABLE ACTUAL_INPUTS (
  ACTUAL_INPUT_ID   OID DEFAULT nextval('ACTUAL_INPUT_SEQ') PRIMARY KEY,
  MODULE_EXECUTION_ID       OID NOT NULL REFERENCES MODULE_EXECUTIONS
                    DEFERRABLE INITIALLY DEFERRED,
  FORMAL_INPUT_ID   OID NOT NULL REFERENCES FORMAL_INPUTS
                    DEFERRABLE INITIALLY DEFERRED,
  INPUT_MODULE_EXECUTION_ID OID NOT NULL REFERENCES MODULE_EXECUTIONS
                    DEFERRABLE INITIALLY DEFERRED
--  ACTUAL_OUTPUT_ID  OID NOT NULL REFERENCES ACTUAL_OUTPUTS
--                    DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX ACTUAL_INPUTS_MODULE_EXECUTION_ID ON ACTUAL_INPUTS(MODULE_EXECUTION_ID);
CREATE INDEX ACTUAL_INPUTS_FI_ID ON ACTUAL_INPUTS(FORMAL_INPUT_ID);
CREATE INDEX ACTUAL_INPUTS_IA_ID ON ACTUAL_INPUTS(INPUT_MODULE_EXECUTION_ID);
COMMENT ON TABLE ACTUAL_INPUTS IS
'This table stores the actual inputs that were used by a specific \
analysis by linking the data type declaration (FORMAL_INPUT_ID) and \
the MODULE_EXECUTION_ID to an INPUT_MODULE_EXECUTION_ID.  All of the outputs of the \
INPUT_ANALYSIS, of the same type as the specified FORMAL_INPUT, are \
considered to be inputs of the specified ANALYSIS.';

CREATE SEQUENCE SEMANTIC_TYPE_OUTPUT_SEQ;
CREATE TABLE SEMANTIC_TYPE_OUTPUTS (
  SEMANTIC_TYPE_OUTPUT_ID   OID DEFAULT NEXTVAL('SEMANTIC_TYPE_OUTPUT_SEQ')
                            PRIMARY KEY,
  MODULE_EXECUTION_ID       OID NOT NULL REFERENCES MODULE_EXECUTIONS
                            DEFERRABLE INITIALLY DEFERRED,
  SEMANTIC_TYPE_ID          OID NOT NULL REFERENCES SEMANTIC_TYPES
                            DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE SEMANTIC_TYPE_OUTPUTS IS
'Records which semantic types a module execution creates attributes of.  \
This allows the analysis engine to determine which semantic types were \
placed into untyped outputs.  Only semantic types which do not have an \
output in the corresponding module are recorded here, to eliminate some \
redundancy of information stored.';

CREATE SEQUENCE ANALYSIS_CHAIN_EXECUTION_SEQ;
CREATE TABLE ANALYSIS_CHAIN_EXECUTIONS (
  ANALYSIS_CHAIN_EXECUTION_ID  OID
                         DEFAULT nextval('ANALYSIS_CHAIN_EXECUTION_SEQ')
                         PRIMARY KEY,
  ANALYSIS_CHAIN_ID       OID NOT NULL REFERENCES ANALYSIS_CHAINS
                         DEFERRABLE INITIALLY DEFERRED,
  DATASET_ID             OID NOT NULL REFERENCES DATASETS
                         DEFERRABLE INITIALLY DEFERRED,
  EXPERIMENTER_ID        OID NOT NULL
                         REFERENCES EXPERIMENTERS
                         DEFERRABLE INITIALLY DEFERRED,
  TIMESTAMP              TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);
CREATE INDEX ANALYSIS_CHAIN_EXECUTIONS_VIEW ON ANALYSIS_CHAIN_EXECUTIONS(ANALYSIS_CHAIN_ID);
CREATE INDEX ANALYSIS_CHAIN_EXECUTIONS_DATASET ON ANALYSIS_CHAIN_EXECUTIONS(DATASET_ID);
CREATE INDEX ANALYSIS_CHAIN_EXECUTIONS_EXPERIMENTER ON ANALYSIS_CHAIN_EXECUTIONS(EXPERIMENTER_ID);
COMMENT ON TABLE ANALYSIS_CHAIN_EXECUTIONS IS
'There is exactly one row in this table for each time an analysis
chain is executed against a dataset.  This information cannot
necessarily be reconstructed from the actual inputs and outputs and
attributes, since there could be attribute reuse involved.';


CREATE SEQUENCE ANALYSIS_NODE_EXECUTION_SEQ;
CREATE TABLE ANALYSIS_NODE_EXECUTIONS (
  ANALYSIS_NODE_EXECUTION_ID  OID 
                              DEFAULT nextval('ANALYSIS_NODE_EXECUTION_SEQ')
                              PRIMARY KEY,
  ANALYSIS_CHAIN_EXECUTION_ID       OID NOT NULL REFERENCES ANALYSIS_CHAIN_EXECUTIONS
                              DEFERRABLE INITIALLY DEFERRED,
  ANALYSIS_CHAIN_NODE_ID       OID NOT NULL REFERENCES ANALYSIS_CHAIN_NODES
                              DEFERRABLE INITIALLY DEFERRED,
  MODULE_EXECUTION_ID                 OID NOT NULL REFERENCES MODULE_EXECUTIONS
                              DEFERRABLE INITIALLY DEFERRED
);
COMMENT ON TABLE ANALYSIS_NODE_EXECUTIONS IS
'This table maps rows in the MODULE_EXECUTIONS table to the nodes which
executed them.  Every time a node is executed in a chain, a row is
created in this table, even if data reuse occurs.';

