-- 030-FeatureAttributes.sql
-- ---------------------------------------
-- Author:  Ilya G. Goldberg (igg@mit.edu)
-- Copyright 1999-2002 Ilya G. Goldberg
-- This file is part of OME.
-- ---------------------------------------
-- 
--     OME is free software; you can redistribute it and/or modify
--     it under the terms of the GNU General Public License as published by
--     the Free Software Foundation; either version 2 of the License, or
--     (at your option) any later version.
-- 
--     OME is distributed in the hope that it will be useful,
--     but WITHOUT ANY WARRANTY; without even the implied warranty of
--     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--     GNU General Public License for more details.
-- 
--     You should have received a copy of the GNU General Public License
--     along with OME; if not, write to the Free Software
--     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-- 


--
--

-- The FEATURES table is used for ALL image features reguardless of
-- algorithm.  The properties of a feature are that it has a variable
-- list of attributes.
-- Feature attribute tables contain a column named feature_id
-- that refers to the attribute_id column in this table.
-- This table is technically an image attribute table because it contains an image_id.
-- Algorithms that define features write rows into this table - one row per feature.
--


-- *** DC 01/24/2003
-- Feature tags in brackets are reserved by OME.
-- Currently used special tags:
--    [Image]
--    [Feature]
--    [Parent]
--    [Iterator]

CREATE SEQUENCE FEATURE_SEQ;
CREATE TABLE FEATURES (
  FEATURE_ID         OID DEFAULT NEXTVAL('FEATURE_SEQ') PRIMARY KEY,
  IMAGE_ID           OID NOT NULL REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
  PARENT_FEATURE_ID  OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
  TAG                VARCHAR(128) NOT NULL,
  NAME               VARCHAR(128)
);
CREATE INDEX FEATURES_IMAGE ON FEATURES(IMAGE_ID);
CREATE INDEX FEATURES_PARENT ON FEATURES(PARENT_FEATURE_ID);
CREATE INDEX FEATURES_TAG ON FEATURES(TAG);
COMMENT ON TABLE FEATURES IS
'The FEATURES table is used for ALL image features reguardless of \
algorithm.  Image features are understood to be objects of interest \
within the image - blobs, cells, nuclei, etc.  The properties of a \
feature are that it has a variable list of attributes.  The kinds of \
attributes in this list and their values define everything known about \
the feature.  Feature attribute tables contain a column named \
FEATURE_ID that refers to the FEATURE_ID column of this table.  \
Algorithms that define features (segmentation algorithms) write rows \
into this table - one row per feature. \
\
Further, features are given tags and arranged hierarchically within an \
image.  Modules can then request to iterate over features at an \
arbitrary point in the hierarchy by specifying a feature tag.  This \
hierarchy is created automatically; new features from a module are \
considered to be children of the iterated feature for that module.';

COMMENT ON COLUMN FEATURES.TAG IS
'This column is a semantic tag assigned to the feature, e.g., "CELL", \
"NUCLEUS", "CHROMOSOME".';




-- DC - March 4, 2003
-- _All_ of the following tables will eventually be replaced with XML
-- descriptors.


-- CREATE TABLE LOCATION (  -- Sorry, but POSITION is a reserved word in SQL, apparently.
--         ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--         ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--         FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--         X             FLOAT,
--         Y             FLOAT,
--         Z             FLOAT
-- );
-- CREATE INDEX LOCATION_ATTRIBUTE_ID_idx ON LOCATION (ATTRIBUTE_ID);
-- CREATE INDEX LOCATION_FEATURE_ID_idx  ON LOCATION (FEATURE_ID);
-- COMMENT ON TABLE LOCATION IS
-- 'The LOCATION table stores positional information (in pixel units) about a Feature';
-- COMMENT ON COLUMN LOCATION.X IS
-- 'Center of volume - not mass';
-- COMMENT ON COLUMN LOCATION.Y IS
-- 'Center of volume - not mass';
-- COMMENT ON COLUMN LOCATION.Z IS
-- 'Center of volume - not mass';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','LOCATION','Spatial locations');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'Z',NULL,'');

--
-- CREATE TABLE SIGNAL (
--         ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--         ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--         FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--         WAVELENGTH       SMALLINT,
--         CENTROID_X      FLOAT,
--         CENTROID_Y      FLOAT,
--         CENTROID_Z      FLOAT,
--         INTEGRAL        FLOAT,
--         MEAN            FLOAT,
--         GEOMEAN         FLOAT,
--         SIGMA           FLOAT,
--         BACKGROUND      FLOAT
-- );
-- CREATE INDEX SIGNAL_ATTRIBUTE_ID_idx ON SIGNAL (ATTRIBUTE_ID);
-- CREATE INDEX SIGNAL_FEATURE_ID_idx  ON SIGNAL (FEATURE_ID);
-- COMMENT ON TABLE SIGNAL IS
-- 'The SIGNAL table stores spectral information about a Feature';
-- COMMENT ON COLUMN SIGNAL.CENTROID_X IS
-- 'Center of mass - not volume';
-- COMMENT ON COLUMN SIGNAL.CENTROID_Y IS
-- 'Center of mass - not volume';
-- COMMENT ON COLUMN SIGNAL.CENTROID_Z IS
-- 'Center of mass - not volume';
-- COMMENT ON COLUMN SIGNAL.INTEGRAL IS
-- 'Sum of all pixel intensities at the specified wavelength for this Feature';
-- COMMENT ON COLUMN SIGNAL.BACKGROUND IS
-- 'The average "background" signal intensity for this Feature';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','SIGNAL','Signals');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'WAVELENGTH',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'CENTROID_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'CENTROID_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'CENTROID_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'INTEGRAL',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MEAN',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'GEOMEAN',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'SIGMA',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'BACKGROUND',NULL,'');


-- CREATE TABLE TIMEPOINT (
--         ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--         ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--         FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--         TIMEPOINT     INTEGER
-- );
-- CREATE INDEX TIMEPOINT_ATTRIBUTE_ID_idx ON TIMEPOINT (ATTRIBUTE_ID);
-- CREATE INDEX TIMEPOINT_FEATURE_ID_idx  ON TIMEPOINT (FEATURE_ID);
-- COMMENT ON TABLE TIMEPOINT IS
-- 'This simply stores the timepoint within the image that this Feature was found.';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','TIMEPOINT','Timepoints');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'TIMEPOINT',NULL,'');


-- CREATE SEQUENCE TRAJECTORY_ID_SEQ;
-- CREATE TABLE DELTA (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   MOVED_TO      OID,
--   DELTA_X       FLOAT,
--   DELTA_Y       FLOAT,
--   DELTA_Z       FLOAT,
--   DISTANCE      FLOAT,
--   VELOCITY      FLOAT,
--   TRAJECTORY_ID OID
-- );
-- CREATE INDEX DELTA_ATTRIBUTE_ID_idx ON DELTA (ATTRIBUTE_ID);
-- CREATE INDEX DELTA_FEATURE_ID_idx  ON DELTA (FEATURE_ID);
-- COMMENT ON TABLE DELTA IS
-- 'This table stores tracking information as Features move through timepoints.';
-- COMMENT ON COLUMN DELTA.DELTA_X IS
-- 'In pixel coordinates.';
-- COMMENT ON COLUMN DELTA.DELTA_Y IS
-- 'In pixel coordinates.';
-- COMMENT ON COLUMN DELTA.DELTA_Z IS
-- 'In pixel coordinates.';
-- COMMENT ON COLUMN DELTA.DISTANCE IS
-- 'In microns - to account for anisotropic 3-D space.';
-- COMMENT ON COLUMN DELTA.VELOCITY IS
-- 'In microns per second.';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','DELTA','Timepoints');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MOVED_TO',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DISTANCE',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'VELOCITY',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'TRAJECTORY_ID',NULL,'');


--
-- CREATE TABLE EXTENT (
--         ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--         ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--         FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--         VOLUME        INTEGER,
--         MIN_X         SMALLINT,
--         MIN_Y         SMALLINT,
--         MIN_Z         SMALLINT,
--         MAX_X         SMALLINT,
--         MAX_Y         SMALLINT,
--         MAX_Z         SMALLINT,
--         SIGMA_X       FLOAT,
--         SIGMA_Y       FLOAT,
--         SIGMA_Z       FLOAT,
--         SURFACE_AREA  FLOAT,    -- For 2D spots, this is the surface area including fenestrations.
--         PERIMITER     FLOAT,    -- The outside perimiter - this will not include inside perimiters in fenestrated spots
--         FORM_FACTOR   FLOAT     -- surface area/volume (3D) or perimiter/surface area (2D) 1.0 for circle or sphere.
-- );
-- CREATE INDEX EXTENT_ATTRIBUTE_ID_idx ON EXTENT (ATTRIBUTE_ID);
-- CREATE INDEX EXTENT_FEATURE_ID_idx  ON EXTENT (FEATURE_ID);

-- COMMENT ON TABLE EXTENT IS
-- 'This table stores shape information about a Feature.';
-- COMMENT ON COLUMN EXTENT.SURFACE_AREA IS
-- 'For 2D spots, this is the surface area (in pixels) including fenestrations.';
-- COMMENT ON COLUMN EXTENT.PERIMITER IS
-- 'The outside perimiter - this will not include inside perimiters in fenestrated spots.';
-- COMMENT ON COLUMN EXTENT.FORM_FACTOR IS
-- 'surface area/volume (3D) or perimiter/surface area (2D) 1.0 for circle or sphere';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','EXTENT','Extents');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'VOLUME',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MIN_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MIN_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MIN_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MAX_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MAX_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'MAX_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'SIGMA_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'SIGMA_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'SIGMA_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'SURFACE_AREA',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'PERIMITER',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'FORM_FACTOR',NULL,'');


-- CREATE TABLE THRESHOLD (
--         ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--         ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--         FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--         THRESHOLD     SMALLINT
-- );
-- CREATE INDEX THRESHOLD_ATTRIBUTE_ID_idx ON THRESHOLD (ATTRIBUTE_ID);
-- CREATE INDEX THRESHOLD_FEATURE_ID_idx  ON THRESHOLD (FEATURE_ID);

-- COMMENT ON TABLE THRESHOLD IS
-- 'This table stores the threshold used for a particular feature.';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','THRESHOLD','Thresholds');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'THRESHOLD',NULL,'');


-- CREATE TABLE ENVELOPE (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   X             SMALLINT,
--   Y             SMALLINT,
--   Z             SMALLINT
-- );
-- CREATE INDEX ENVELOPE_ATTRIBUTE_ID_idx ON ENVELOPE (ATTRIBUTE_ID);
-- CREATE INDEX ENVELOPE_FEATURE_ID_idx  ON ENVELOPE (FEATURE_ID);

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','ENVELOPE','Envelopes');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'Z',NULL,'');


-- CREATE TABLE DISTANCE_TO_SPINDLE_AXIS (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   DISTANCE      FLOAT,
--   LENGTH        FLOAT
-- );
-- CREATE INDEX DISTANCE_TO_SPINDLE_AXIS_ATTRIBUTE_ID_idx ON DISTANCE_TO_SPINDLE_AXIS (ATTRIBUTE_ID);
-- CREATE INDEX DISTANCE_TO_SPINDLE_AXIS_FEATURE_ID_idx  ON DISTANCE_TO_SPINDLE_AXIS (FEATURE_ID);

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','DISTANCE_TO_SPINDLE_AXIS','Distances to spindle axis');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DISTANCE',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'LENGTH',NULL,'');


-- CREATE SEQUENCE CLUSTER_ID;
-- CREATE TABLE CLUSTERS (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   CLUSTER_ID    OID
-- );
-- CREATE INDEX CLUSTERS_ATTRIBUTE_ID_idx ON CLUSTERS (ATTRIBUTE_ID);
-- CREATE INDEX CLUSTERS_FEATURE_ID_idx  ON CLUSTERS (FEATURE_ID);

-- COMMENT ON TABLE CLUSTERS IS
-- 'This table allows features to be assigned to clusters.  Features may belong to more than one cluster.';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','CLUSTERS','Clusters');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'CLUSTER_ID',NULL,'');


-- CREATE TABLE NEAREST_NEIGHBOR (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   NEIGHBOR_ID    OID,
--   DELTA_X       FLOAT,
--   DELTA_Y       FLOAT,
--   DELTA_Z       FLOAT,
--   DISTANCE      FLOAT
-- );
-- CREATE INDEX NEAREST_NEIGHBOR_ATTRIBUTE_ID_idx ON NEAREST_NEIGHBOR (ATTRIBUTE_ID);
-- CREATE INDEX NEAREST_NEIGHBOR_FEATURE_ID_idx  ON NEAREST_NEIGHBOR (FEATURE_ID);

-- COMMENT ON TABLE NEAREST_NEIGHBOR IS
-- 'This table sores some information about a feature\'s nearest neighbor.';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','NEAREST_NEIGHBOR','Nearest neighbors');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'NEIGHBOR_ID',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_X',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_Y',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DELTA_Z',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'DISTANCE',NULL,'');


-- CREATE TABLE BINARY_IMAGE (
--   ATTRIBUTE_ID  OID DEFAULT NEXTVAL('ATTRIBUTE_ID') PRIMARY KEY,
--   ANALYSIS_ID   OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
--   FEATURE_ID    OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
--   IMAGE_TYPE    VARCHAR(32),     -- Should match one of the IMAGE_TYPEs in the IMAGES table.
--   IMAGE_ID_IN   OID,             -- The image id that this is a binary image of.
--   IMAGE_ID_BIN  OID,             -- Does not necessarily point to an actual image ID (could be NULL).
--   NAME          VARCHAR(256),
--   PATH          VARCHAR(256)
-- );

-- CREATE INDEX BINARY_IMAGE_ATTRIBUTE_ID_IDX ON BINARY_IMAGE (ATTRIBUTE_ID);
-- CREATE INDEX BINARY_IMAGE_FEATURE_ID_IDX  ON BINARY_IMAGE (FEATURE_ID);
-- CREATE INDEX BINARY_IMAGE_IMAGE_ID_IDX  ON BINARY_IMAGE (IMAGE_ID_IN);
-- CREATE RULE EXPIRE_BINARY_IMAGE_FILES AS
--   ON DELETE TO BINARY_IMAGE DO
--   INSERT INTO EXPIRED_FILES VALUES (OLD.PATH || OLD.NAME);

-- COMMENT ON TABLE BINARY_IMAGE IS
-- 'This table stores information about binary images (masks).';

-- INSERT INTO DATA_TABLES VALUES 
-- (NEXTVAL('DATA_TABLE_SEQ'),'F','BINARY_IMAGE','Binary images');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'IMAGE_TYPE',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'IMAGE_ID_IN',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'IMAGE_ID_BIN',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'NAME',NULL,'');
-- INSERT INTO DATA_COLUMNS VALUES
-- (NEXTVAL('DATA_COLUMN_SEQ'),CURRVAL('DATA_TABLE_SEQ'),'PATH',NULL,'');
