#!/usr/bin/perl

# bootstrapOME.pl -- OME Database and initial environment script.
#
# Copyright (C) 2002 Open Microscopy Environment
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#*********
#********* INCLUDES
#*********

use warnings;
use strict;
use DBI;
use Sys::Hostname;
use Term::ReadKey;
use MIME::Base64;
use File::Copy;
use Getopt::Long;


# You must install OME before running bootstrap.
use OME::SessionManager;
use OME::Tasks::OMEImport;

use OME::Configuration;
use OME::ImportExport::ChainImport;

#*********
#********* GLOBALS AND DEFINES
#*********

# Default directory locations
my %dirs = (
	baseDir		=> "/OME",
	xmlDir 		=> "/OME/xml",
	binDir 		=> "/OME/bin",
	reposDir 	=> "/OME/repository",
	tmpDir 		=> "/var/tmp/OME",
	lockDir 	=> "/var/tmp/OME/lock",
	sessionDir 	=> "/var/tmp/OME/sessions"
);

# Default database location
my $dataSource =   "dbi:Pg:dbname=ome";

# Checking for the existance of a Debian/RedHat system
# FIXME: We really should have something here for OS X, and the BSD's
my $platform = "UNKNOWN";
if (-e "/etc/debian_version") {
        $platform = "DEBIAN";
} elsif (-e "/etc/redhat_version") {
        $platform = "REDHAT";
}

# Command line options
my ($skipPasswordCheck, $defaultDirectories, $defaultUserDetails, $help, $fast);
GetOptions ("s|skip-password-check", \$skipPasswordCheck,	# Skip the PW check
	    "d|default-directories", \$defaultDirectories,	# Use default OME dirs
	    "n|default-user-details", \$defaultUserDetails,	# Use default user details
	    "f|fast", \$fast,					# Fast operation
	    "h|help", \$help					# Show usage
	    );

# CLI flag logic
if ($fast) { $skipPasswordCheck 	= 1;
	     $defaultDirectories 	= 1;
	     $defaultUserDetails	= 1; }
if ($help) { usage() }


#*********
#********* LOCAL SUBROUTINES
#*********

# Password grab
sub getPassword {
	my ($ok, $input);

	ReadMode(2);
	do {
		print "Password: ";
		$input = ReadLine(0);
		print "\n";
		chomp($input);
		print "Verify Password: ";
		my $input2 = ReadLine(0);
		print "\n";
		chomp($input2);
		if(!$skipPasswordCheck && length($input) < 6 ) {
			print "Password must be at least 6 characters long.\n";
		} elsif( not( $input eq $input2) ) {
			print "Passwords do not match. Please reenter.\n";
		} else {
			$ok = 1;
		}
	} while( not $ok);

	#my $password = $passwordInput;
	my $salt = join('',('.','/',0..9,'A'..'Z','a'..'z')[rand 64, rand 64]);
	my $crypt = crypt($input,$salt);

	ReadMode(1);
	return ($input, $crypt);
}

# Usage display
sub usage {
	my $usage = <<USAGE;
Bootstrap the OME database, create inital user(s) and do environment setup.

Usage:
  $0 [options]

Options:
  -s, --skip-password-check	Skip the password check
  -d, --default-directories	Run default directory creation (UID 0 is
  				required as the OME directories are off
				the root)
  -n, --default-user-details	Use default user details for the 
				creation of the initial OME experimenter
  -f, --fast			Fast operation for developers who know
  				they want the default OME directory
				structure, want default user details and
				want to skip the password check
  -h, --help			This message

Report bugs to <ome-devel\@mit.edu>.
USAGE

	print STDERR $usage;
	exit (0);
}

# Determine host's MAC address
sub getMAC {
    my $macAddr;
    my ($ifinfo, @ifinfo);

# First, what's our OS?
    my $os = `uname -s`;   # assumes POSIX compliant uname cmnd
    chomp $os;
# now get MAC address using OS specific command
    if ($os =~ /Linux/) {
	@ifinfo = `/sbin/ifconfig eth0`;
	$macAddr = $ifinfo[0];
	$macAddr =~ s/^.*HWaddr\s+(.*)$/$1/;
    } elsif ($os =~ /Darwin/) {
	@ifinfo = `/sbin/ifconfig`;
	# Get the first line containing the word 'ether', then set the adress to everything after 'ether'
	foreach (@ifinfo) {
		if ($_ =~ /ether\s+(.*)$/) {$macAddr = $1;last;};
	}
    } elsif ($os =~ /HPUX/) {
	@ifinfo = `lanscan`;
	$macAddr = $ifinfo[2];
	$macAddr =~ s/^.*0x([0-9A-F]+).*$/$1/;
    } elsif ($os =~ /(Solaris)|(SunOS)|(sunos)/) {
	$macAddr = `ifconfig -a`;              # need to su, & then run ifconfig -a & use
	$macAddr =~ s/.*ether: ([^ \t]+)$/$1/  # the colon separated string after 'ether'
    } elsif ($os =~ /FreeBSD/) {
	$macAddr = `dmefg`;
    } elsif ($os =~ /MS-DOS/) {    # good luck running on NT
	$macAddr = `ipconfig \all`;
	$macAddr =~ s/^.*([0-9A-F][0-9A-F]-[0-9A-F][0-9A-F]-[0-9A-F][0-9A-F]-[0-9A-F][0-9A-F]-[0-9A-F][0-9A-F]-[0-9A-F][0-9A-F]).*$/$1/;
    } else {
	$macAddr = "";
    }
    chomp $macAddr;
    #print "$macAddr\n";
    return $macAddr;
}


# Convert a username to a UID
sub user2uid ($)
{
  my $u = shift;
  return $u if (numberp ($u));
  my $uid = getpwnam ($u);
  return $uid if (defined $uid && numberp ($uid));
  print "no such user\n";
  return (undef);
}

# Script implementation of Unix su
sub set_uid ($)
{
  my $suid = user2uid (shift) or return undef;
  my $euid = $>;
  return 1 if $suid == $euid;

  $> = 0;
  die 'You cannot change your UID unless you are root - sorry' if $> != 0;
  $> = $suid;
  die "cannot set uid: $!\n" if ($> == $euid && $euid != $suid);
  return $suid;
  
}

# Get a username
sub get_uid ()
{
my @login = getpwuid($>);
	return ( $login[0] );
}

# Create a postgres superuser (CREATE USER foo CREATEUSER)
sub createPostgresSuperUser ($)
{
my $userName = shift;
my $uid = get_uid ();
my $output;


	eval { set_uid ('postgres') };
	$output = `createuser -d -a $userName 2>&1`;
	die "Couldn't execute createuser (a postgres utility)! Is createuser in your \$PATH?\n"
		unless defined $output and $output;
	if ($output =~ /already exists/) {
		print "User $userName already exists\n";
		undef $userName;
	} elsif ($output =~ /^CREATE/) {
		print "User $userName created\n";
	} else {
		undef $userName;
		die "Error creating user: $output\n";
	}
	eval { set_uid ($uid) };
	return ($userName);
}

sub numberp ($)
{
  return (defined $_[0] && $_[0] =~ /^-?\d+$/o) ? 1 : 0;
}

sub getApacheUser () {
	my $apacheUser;

	# Grab our Apache user from the password file
        open (PW_FILE, "<", "/etc/passwd") or die ("Couldn't open /etc/passwd ", $!, ".\n");
        while (<PW_FILE>) {
                chomp;
                my $username = (split ":")[0];
                if ($username =~ /httpd|apache|www-data|www/) {
                        return $username;
                }
        }
        close (PW_FILE);
                                                                                                 
        # We couldn't get the username from the password file so lets ask for it
        print "Could not determine Apache user.\nWhat is the unix name that Apache runs under? ";
	my $apacheUserIn = <STDIN>;
        chomp $apacheUserIn;
        if ($apacheUserIn and user2uid ($apacheUserIn)) {
                $apacheUser = $apacheUserIn;
        } else { die "User ", $apacheUserIn, "does not exist.\n" };
                                                                                                 
        return $apacheUser;
}

sub getApacheConf () {

	# First, get the httpd executable.
	my $httpdBin;
	foreach (split (' ',`whereis httpd`)) {
			chomp;
			if (-e $_ and not -d $_ and -x $_ and -B $_) {
				$httpdBin = $_;
				last;
			}
		}
	print "httpd bin is '$httpdBin'\n";
	
	# Get the location of httpd.conf from the compiled-in options to httpd
	my $httpdConf = `$httpdBin -V | grep SERVER_CONFIG_FILE | cut -d '"' -f 2`;
	chomp $httpdConf;
	if (not -e $httpdConf) {
		# on some systems the path reported is not absolute, but relative to HTTPD_ROOT.
		my $httpdRoot = `$httpdBin -V | grep HTTPD_ROOT | cut -d '"' -f 2`;
		chomp $httpdRoot;
		$httpdConf = File::Spec->catpath ($httpdRoot,$httpdConf);
	}
	
	# If we still haven't found it, ask the user.
	if (not -e $httpdConf or not -f $httpdConf) {
		$httpdConf = '/etc/httpd/conf/httpd.conf';
		print "Location of httpd.conf [$httpdConf]: ";
		my $omeInput = <STDIN>;
		chomp($omeInput);
		$httpdConf = $omeInput if $omeInput;
	}
	
	# If we still haven't found it, die.
	die "Could not find httpd.conf file\n" unless -e $httpdConf and -f $httpdConf;
	
	print "httpd conf is '$httpdConf'\n";
}

#*********
#********* START OF CODE
#*********

print "\nDetected the platform to be: ", $platform, "\n";

# global to hold combined STDOUT and STDERR for backticks commands
my $CMD_OUT;

# Bold text surrounded by "\033[1m" and "\033[0m"
print "\033[1m\n----------------------------\n";
print " OME base directories setup\n";
print "----------------------------\033[0m\n";

# We need to drop our umask so that everyone can create files.
print "Dropping umask to \033[1m\"0000\"\033[0m.\n";
umask (0000);

# Check if we're using the default directories
if (! $defaultDirectories) {
	# Okay, we're not so lets ask for them
	print "\nI will now ask you for the locations for certain OME base directories. \n";
	print "Pressing [Enter] for any question will select the standard base directory.\n\n";
	ReadMode(1);

	# baseDir
	print "Base OME directory [/OME]: ";
	my $basedirIn = <STDIN>;
	chomp $basedirIn;
	if ($basedirIn) {
       		$dirs{baseDir} = $basedirIn;
	}

	# reposDir
	print "Base respository directory [$dirs{reposDir}]: ";
	my $reposdirIn = <STDIN>;
	chomp $reposdirIn;
	if ($reposdirIn) {
	        $dirs{reposDir}= $reposdirIn;
	}

	# tmpDir
	print "Base tmp directory [$dirs{tmpDir}]: ";
	my $tmpdirIn = <STDIN>;
	chomp $tmpdirIn;
	if ($tmpdirIn) {
		$dirs{tmpDir} = $tmpdirIn;
	}

	print "\n";  # Spacing
} elsif (($defaultDirectories) && ($> != 0)) {
	# We're using the defaults but aren't root
	die ("You have chosen to use the default OME directories but you are not root!\n");
}

# Create necessary directories, default or user supplied
print "\nCreating necessary directories...\n";
foreach my $dir (sort (values (%dirs))) {
	if (not -e $dir) {
		# Bold text surrounded by "\033[1m" and "\033[0m"
		print ("\033[1m\tCreating \"", $dir, "\".\033[0m\n");
		mkdir ($dir, 0777) or die ("Couldn't make directory ", $dir, ". ", $!, ".\n");
       	}
}
print "\t\033[1m[Done.]\033[0m\n";

# Stylesheet copies using File::Copy (much cleaner)
print "Copying stylesheets...\n";
my @files = glob ("../xml/xslt/*.xslt");
foreach my $file (@files) {
	print "\t\033[1m$file\033[0m\n";
        copy ($file, $dirs{xmlDir}) or die ("Couldn't copy file ", $file, ". ", $!, ".\n");
}
print "\t\033[1m[Done.]\033[0m\n";

# Default LSID authority
my $LSIDauthority = hostname;
$LSIDauthority =~ s/^\W*\s*//;
$LSIDauthority =~ s/\W*\s*$//;

# Grab an LSID from the user if they so desire
print "LSID Authority [$LSIDauthority]: ";
my $LSIDauthorityIn = <STDIN>;
chomp $LSIDauthorityIn;
if ($LSIDauthorityIn) {
	$LSIDauthority = $LSIDauthorityIn;
}

# We need an OME administrator
if ($> == 0) { print "You are logged in as root.  You will need to assign an OME admin user.\n" }

# Grab our user
my $adminUser = get_uid();

print "User to administer OME [$adminUser]: ";
my $adminUserIn = <STDIN>;
chomp $adminUserIn;
if ($adminUserIn) {
        $adminUser = $adminUserIn;
}

set_uid ($adminUser) or die "The admin user must exist on the system\n";

my $apacheUser = getApacheUser();
createPostgresSuperUser ($adminUser);
createPostgresSuperUser ($apacheUser);

#
# IGG:  Could be a good idea to put in some code here to determine if the database is already up
# This would require DB versioning to make sure the DB version is the same as the API's
# Possibly OME::DBConnection::DBversion compared with CONFIGURATION.DB_VERSION 
print "\033[1m\n----------------------------------\n";
print " OME database schema bootstrapper\n";
print "----------------------------------\033[0m\n\n";

print "Creating PostgreSQL database...\n";

# Connect to the template DB so we can create our OME database
my $dbh = DBI->connect("dbi:Pg:dbname=template1");
die ($dbh->errstr) unless $dbh;

$dbh->do ("CREATE DATABASE ome") or die ($dbh->errstr);
print "\t\033[1m[Done.]\033[0m\n";

# Disconnect our temporary database handle
$dbh->disconnect;

# Debian does this by default using Debconf, everything else needs it.
# Atleast as far as I know. (Chris)
if ($platform ne "DEBIAN") {
        print "Adding PL-PGSQL language...\n";
        $CMD_OUT = `createlang plpgsql ome 2>&1`;
        die $CMD_OUT if $? != 0;
	print "\t\033[1m[Done.]\033[0m\n";
}

$dbh = DBI->connect($dataSource); 
die ($dbh->errstr) unless $dbh;

# Fixes a problem in PostgreSQL 7.2 where the OID type is no longer type-
# compatible with sequences, preventing their use as primary keys.
#
# Merged in from oid-fix.sql

print "Fixing OID/INTEGER compatability bug...\n";

my $sql = <<SQL;
CREATE FUNCTION OID(INT8) RETURNS OID AS '
declare
  i8 alias for \$1;
begin
  return int4(i8);
end;'
LANGUAGE 'plpgsql';
SQL

$dbh->do($sql) or die ($dbh->errstr);
print "\t\033[1m[Done.]\033[0m\n";

print "Loading the database schema...\n";

# Glob all the SQL files in the CWD and process each in a sorted fashion into the DB
@files = glob("*.sql");
foreach my $file (sort @files) {
	if ($file =~ /^[0-9]{3}-/) {
		print "\t$file\n";
		# Apparently psql doesn't return non-zero error status if there were problems
		# processing the files. So, we have to process the lines one by one and look
		# for 'ERROR:'
		my @OUTPUT = `psql -d ome -q -f $file  2>&1`;
		my $errStat;
		foreach (@OUTPUT) {
			next unless /ERROR:/;
			print;
			die "Problems importing $file - Database creation failed!\n";
		}
	}
}
print "\t\033[1m[Done.]\033[0m\n";

# [0-8] -- username, pw_hash, uid, gid, comment, gcos, name, homedir, shell 
# [0,1,6,7] -- username, pw_hash, name, homedir
my ($omeName, $password, $personalInfo, $dataDir) = (getpwuid($>))[0,1,6,7];
my $fullName = (split (',',$personalInfo,2))[0];
my ($firstName,$lastName) = split (' ',$fullName,2);
my $email = $omeName.'@'.hostname;


# Find out our MAC address. This will be put into the Configuration table,
# and used at image creation time to form part of each Image's GUID.
my $macAddr = getMAC();

# The DB instance uniquely identifies this specific DB instance on this machine
# Can only make one per second - sorry.
# There are two punctuation characters in base64: '/' and '+' (no '=' here b/c of trunctation)
# substitute '/' with something safer - '_'.
my $DBinstance = substr (encode_base64(pack ('N',time()),''),0,6);
$DBinstance =~ s/\//_/g;

# Bold text surrounded by "\033[1m" and "\033[0m"
print "\n\033[1mCreating initial OME user...\033[0m\n\n";

# User details:
# $personalInfo, $omeName, $passwordInput, $email, $datadir
my ($passwordInput, $cryptPassword);  # Need this declaration here for scope.
if (! $defaultUserDetails) {
	print "I will now ask you for a username and password for the\n";
	print "new OME user.  Pressing [Enter] for either will cause\n";
	print "OME to use the same value as your Unix login.\n\n";


	ReadMode(1);
	print "Full name (first last) [$firstName $lastName]: ";
	my $personalInfoIn = <STDIN>;
	chomp($personalInfoIn);
	if ($personalInfoIn) {
		$personalInfo = $personalInfoIn if $personalInfoIn;
		$fullName = (split (',',$personalInfo,2))[0];
		($firstName,$lastName) = split (' ',$fullName,2);
	}

	print "Username [$omeName]: ";
	my $omeInput = <STDIN>;
	chomp($omeInput);
	$omeName = $omeInput if $omeInput;

	($passwordInput, $cryptPassword) = getPassword ();	

	ReadMode(1);
	print "email [$email]: ";
	my $emailInput = <STDIN>;
	chomp($emailInput);
	$email = $emailInput if $emailInput;


	print "Data directory [$dataDir]: ";
	my $dataDirInput = <STDIN>;
	chomp ($dataDirInput);
	$dataDir = $dataDirInput if $dataDirInput;
	if (not -e $dataDir) {
		print "Directory '$dataDir' does not exist.  Create it? [No]: ";
		my $YorN = <STDIN>;
		$YorN = uc (chomp($YorN));
		if ($YorN && ($YorN eq 'Y' || $YorN eq 'YES')){
			die "Couldn't make directory $dataDir: $!\n" unless mkdir ($dataDir,0755);
		}
	}
} else {
	print "I will now ask you for a password for the new OME user:\n";
	($passwordInput, $cryptPassword) = getPassword ();
}
	

print "\n";

print "Performing core database operations...\n";

print "\t\033[1mAdding experimenter.\033[0m\n";

$sql = <<SQL;
INSERT INTO experimenters (
	ome_name,
	firstname,
	lastname,
	email,
	password,
	data_dir
	) VALUES (
	'$omeName',
	'$firstName',
	'$lastName',
	'$email',
	'$cryptPassword',
	'$dataDir');
SQL
$dbh->do($sql) or die ($dbh->errstr);

# We don't need our DBH any more so lets be nice and disconnect
$dbh->disconnect;

print "\t\033[1mLogging into OME.\033[0m\n";
my $manager = OME::SessionManager->new();
my $session = $manager->createSession($omeName,$passwordInput);
my $factory = $session->Factory();


print "\t\033[1mAdding configuration entry.\033[0m\n";
my $configuration = $factory->
  newObject("OME::Configuration",
            {
             configuration_id => 1,
             mac_address      => $macAddr,
             db_instance      => $DBinstance,
             lsid_authority   => $LSIDauthority,
             tmp_dir          => $dirs{tmpDir},
             xml_dir          => $dirs{xmlDir},
             bin_dir          => $dirs{binDir},
             repository_dir   => $dirs{reposDir},
             ome_root         => $dirs{baseDir}
            });
print "\t\033[1m[Done.]\033[0m\n";

# XML installation file

my $omeImport = OME::Tasks::OMEImport->
  new(
      session => $session,
      debug => $ENV{OME_IMPORT_DEBUG},
     );

print "Loading OME core XML...\n";

# get list of files
open( CORE_XML, "< CoreXML" )
	or die "Could not open file 'CoreXML'";
my @coreXML;
while( <CORE_XML> ) { chomp($_); push( @coreXML, $_ ) if /^[^#]/; }
close ( CORE_XML );

foreach my $filename (@coreXML) {
    print "$filename\n";
    eval {
        $omeImport->importFile("../xml/$filename",
                               NoDuplicates           => 1,
                               IgnoreAlterTableErrors => 1);
    };

    die "Error when importing $filename\nError message:\n$@"
      if $@;
    $session->commitTransaction();
}

my $experimenter = ($factory->
  findAttributes("Experimenter", undef))[0]
  or die "Could not load Experimenter STs\n";

print "Adding group...\n";
my $group = $factory->
  newAttribute("Group",undef,undef,
               {
                Name    => 'OME',
                Leader  => $experimenter->id(),
                Contact => $experimenter->id()
               });


print "Adding experimenter to group...\n";
$experimenter->Group($group->id());
$experimenter->writeObject();


print "Creating repository entry...\n";
if ($dirs{reposDir} !~ m|.*/$|) {
    $dirs{reposDir} .= "/";
}

my $repository = $factory->
  newAttribute("Repository",undef,undef,
               {
                Path => $dirs{reposDir}
               });

print "Committing changes...\n";
$repository->writeObject();


# Image import chain creation added by Josiah (siah@nih.gov) on March 4, 2003
print "Installing core analysis chains...\n";

# get list of files
open( CORE_CHAINS, "< CoreChains" )
	or die "Could not open file 'CoreChains'";
my @chains;
while( <CORE_CHAINS> ) { chomp($_); push( @chains, $_ ) if /^[^#]/; }
close ( CORE_CHAINS );

my $chainImport = OME::ImportExport::ChainImport->
  new(session => $session);

foreach (@chains) {
    print "$_\n";
    eval {
        $chainImport->importFile("../xml/$_",NoDuplicates => 1);
    };
    die "Error when importing $_\nError message:\n$@\n" if $@;
    $session->DBH()->commit();
}

print "Finding image import module and chain.\n";


# There should be only one of each of these.  If there's more than
# one, we don't care which.  (Maybe we should throw an error instead?)

my $importModule = $factory->
  findObject("OME::Module",name => 'Importer');
$configuration->import_module($importModule);

my $importChain = $factory->
  findObject("OME::AnalysisChain",name => 'Image import analyses');
$configuration->import_chain($importChain);

$configuration->writeObject();


print "Logging out...\n";

$manager->logout($session);

print "\033[1m\nDone. The OME database should be set up.\033[0m\n\n";

# hack added by josiah 6/9/03
# this adds a column (PIXELS_ID) to IMAGES table for the purpose of storing the
# "default pixels" associated with the image.
	my $file = "ImageHack.sql";
#	print "  $file...\n";
	# Apparently psql doesn't return non-zero error status if there were problems processing the files.
	# So, we have to process the lines one by one and look for 'ERROR:'
	my @OUTPUT = `psql -d ome -q -f $file  2>&1`;
	my $errStat;
	foreach (@OUTPUT) {
		next unless /ERROR:/;
		print;
		die "Problems importing $file - Database creation failed!\n";
	}
# end hack added by josiah 6/9/03



