-- Author:  Ilya G. Goldberg (igg@mit.edu)
-- Copyright 1999-2002 Ilya G. Goldberg
-- This file is part of OME.
-- 
--     OME is free software; you can redistribute it and/or modify
--     it under the terms of the GNU General Public License as published by
--     the Free Software Foundation; either version 2 of the License, or
--     (at your option) any later version.
-- 
--     OME is distributed in the hope that it will be useful,
--     but WITHOUT ANY WARRANTY; without even the implied warranty of
--     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--     GNU General Public License for more details.
-- 
--     You should have received a copy of the GNU General Public License
--     along with OME; if not, write to the Free Software
--     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-- 
--
-- The first tables (INSTRUMENTS, PROJECTS, IMAGES, ANALYSES, FEATURES, PROGRAMS, ATTRIBUTE_LIST) are
-- used by all components of OME.  Obviously a given component doesn't have to use all
-- of them.  The tables that follow are specific to the first analysis functions of OME,
-- mainly findSpots and its derivatives (i.e. trackSpots).
-- 
-- Note that the first set of tables is quite primite right now.  The layout is mainly to
-- contain some basic information and to specify the relationships.  In the future, many more columns
-- will need to be added to these tables, and possibly additional tables I haven't thought of.
-- 
-- Some data type conventions:
-- Coordinates:  Dataset coordinates are expressed as 2-byte integers, or 4-byte floating
-- points in the case of centroids or other quantities that are the result of averages or other calculations.
-- Wavelegths are 2-byte integers and expressed in nm (maximum wavelegth is 32767 nm or 32 um)
-- The signal intensity is assumed to be no greater than 2 billion counts (an INT4).
-- Integrals and other intensity sums are 8-byte integers (INT8).
-- Someday someone will make a camera that has more than 32767 x 32767 pixels,
-- and collects data to better than 31 bits.  That's the day I'll regret this decision.
-- Also, right now there is no support for floating-point pixel values.  In a way, floating
-- point is way over-rated.  There is more than enough dynamic range in a 4-byte integer to express
-- any raw or derived data (more than 9 significant figures) - in the vast majority of cases, there
-- is more than enough in a 1-byte integer as long as the scale is properly selected.
-- To express floating-point pixels, one could simply multiply (or divide) them by an appropriate
-- constant.
-- The datatypes are as defined in PostgreSQL convention.  Someday I'll find a full listing of Oracle
-- datatypes and map them accordingly.  Also, I always try to specify the sizes of the numeric
-- datatypes (INT4 instead of INTEGER) to avoid any confusion, and make implicit numerical limits
-- explicit.
--
-- An additional point.  Primary keys are generated as sequences instead of OIDs.  This makes them
-- more visible to the user at the expense of storage space.  Every row of every table has an OID,
-- which could easily serve as the primary key.  OIDs are not generally displayed as part of column
-- listings unless specifically asked for.  The other disadvantage is that they are non-sequential
-- in a given table.  These disadvantages seem minor, but for now at least, primary keys are generated
-- from sequences specific for a given table.  This may change.
--
CREATE SEQUENCE INSTRUMENT_SEQ;
CREATE TABLE INSTRUMENTS (
-- The primary key is a sequence which is automatically entered by default.
INSTRUMENT_ID      OID   DEFAULT nextval ('INSTRUMENT_SEQ') PRIMARY KEY,
NAME               VARCHAR(30),
DESCRIPTION        TEXT
);
--
CREATE SEQUENCE EXPERIMENTER_SEQ;
CREATE TABLE EXPERIMENTERS (
EXPERIMENTER_ID    OID   DEFAULT nextval ('EXPERIMENTER_SEQ') PRIMARY KEY,
OME_NAME           VARCHAR(30),
FIRSTNAME          VARCHAR(30),
LASTNAME           VARCHAR(30),
EMAIL              VARCHAR(50),
password           varchar(64),
GROUP_ID           OID   -- This refers to the table groups, which is not yet defined.  Groups in turn has references
                         -- to experimenters, creating a circular reference.  We create the foreign key constraint after
                         -- the experimenters table.
);
--
--  These are research groups that use OME.
--  By default, the Image's group is set to the Experimenter's group at the time of import.
--  The group leader and contact refer to the experimenters table.
CREATE SEQUENCE GROUP_SEQ;
CREATE TABLE GROUPS (
GROUP_ID    OID    DEFAULT nextval ('GROUP_SEQ') PRIMARY KEY,
NAME               VARCHAR(30),
LEADER             OID REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
CONTACT            OID REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED
);
--
ALTER TABLE experimenters ADD CONSTRAINT groupfk FOREIGN KEY (GROUP_ID) REFERENCES GROUPS(GROUP_ID)
        DEFERRABLE INITIALLY DEFERRED;

--
-- IGG added 9/12/02
-- A dataset is an arbitrarily named collection of images.  Datasets have experimenter and group ownership.
-- The experimenter and group are set to the experimenter and experimenter's group at the time of Dataset creation.
-- They don't have to match the image's experimenter and group.
-- A dataset represents a collection of images that are always analyzed as a unit.
-- Images can belong to more than one dataset.
-- Datasets in turn can belong to more than one project.
CREATE SEQUENCE DATASET_SEQ;
CREATE TABLE DATASETS (
       DATASET_ID     OID DEFAULT nextval('DATASET_SEQ') PRIMARY KEY,
       NAME                  VARCHAR(256) NOT NULL,
       OWNER_ID       OID NOT NULL REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
       GROUP_ID       OID REFERENCES GROUPS DEFERRABLE INITIALLY DEFERRED,
       DESCRIPTION    TEXT,
       LOCKED              BOOLEAN NOT NULL DEFAULT FALSE
);

--
-- IGG added 9/12/02
--  A project is an arbitrarily named collection of datasets.
--  Projects have experimenter and group ownership.
CREATE SEQUENCE PROJECT_SEQ;
CREATE TABLE PROJECTS (
       PROJECT_ID         OID   DEFAULT nextval ('PROJECT_SEQ') PRIMARY KEY,
       NAME               VARCHAR(64) NOT NULL,
       OWNER_ID           OID NOT NULL REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
       GROUP_ID           OID REFERENCES GROUPS DEFERRABLE INITIALLY DEFERRED,
       VIEW               VARCHAR(64),  -- The name of the most recently used view for this project.
       DESCRIPTION        TEXT
);

--
-- IGG added 9/12/02
-- Projects and datasets have a many to many relationship represented by this map.
CREATE TABLE PROJECT_DATASET_MAP (
       PROJECT_ID   INTEGER NOT NULL REFERENCES PROJECTS DEFERRABLE INITIALLY DEFERRED,
       DATASET_ID   INTEGER NOT NULL REFERENCES DATASETS DEFERRABLE INITIALLY DEFERRED,
       PRIMARY KEY (PROJECT_ID, DATASET_ID)
);



--
--  These are display settings for images.  Each image points to a default setting.
--  Also, each user session has a default setting.
--  By default, the image setting over-rides the session setting.
--  There is also a system setting which is over-ridden by any session or image settings.
--  This is mean to control the viewer, but also serves as a mechanism for generating a thumbnail
--  in a non-interactive fashion.
--  The viewer is supposed to allow the user to set his session setting, the image's setting, etc.
CREATE SEQUENCE IMAGE_DISPLAY_SETTINGS_SEQ;
CREATE TABLE IMAGE_DISPLAY_SETTINGS (
        IMAGE_DISPLAY_SETTINGS_ID     OID    DEFAULT nextval ('IMAGE_DISPLAY_SETTINGS_SEQ') PRIMARY KEY,
        URL                           TEXT --  This is meant to be a GET request to a cgi, but can be static also.
);

--
-- IGG added 9/12/02
-- This maintains a path to a repository-format file - a raw pixel dump.
CREATE SEQUENCE REPOSITORY_SEQ;
CREATE TABLE REPOSITORIES (
       REPOSITORY_ID          OID DEFAULT nextval('REPOSITORY_SEQ') PRIMARY KEY,
       PATH                          VARCHAR(256) NOT NULL UNIQUE
);
--
-- This is the main image object, which may be composed of several files.
-- A full image description make use of several auxilliary tables listed below.
CREATE SEQUENCE IMAGE_SEQ;
CREATE TABLE IMAGES (
        IMAGE_ID         OID DEFAULT nextval('IMAGE_SEQ') PRIMARY KEY,
        IMAGE_GUID       VARCHAR(256),     --  See the OME Image GUID Spec
        NAME                VARCHAR(256) NOT NULL,
        DISPLAY_SETTINGS OID REFERENCES image_display_settings DEFERRABLE INITIALLY DEFERRED,
        DESCRIPTION      TEXT,        
        INSTRUMENT_ID    OID REFERENCES INSTRUMENTS DEFERRABLE INITIALLY DEFERRED,
        EXPERIMENTER_ID  OID NOT NULL REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
        GROUP_ID         OID REFERENCES GROUPS,
        CREATED          TIMESTAMP NOT NULL,
        INSERTED         TIMESTAMP NOT NULL,
        IMAGE_TYPE       OID,       -- Circular reference, the fkey will be declared after the IMAGE_TYPES table
        REPOSITORY_ID    OID NOT NULL REFERENCES REPOSITORIES (REPOSITORY_ID) DEFERRABLE INITIALLY DEFERRED,
        PATH             VARCHAR(256) NOT NULL
);
CREATE INDEX images_image_id_idx ON images (image_id);
CREATE INDEX images_name_idx ON images (name);
--
-- IGG added 9/12/02
-- This table is a map between images and datasets.  Images may belong to more than one dataset.
CREATE TABLE IMAGE_DATASET_MAP (
        IMAGE_ID           OID NOT NULL REFERENCES IMAGES (IMAGE_ID) DEFERRABLE INITIALLY DEFERRED,
        DATASET_ID         OID NOT NULL REFERENCES DATASETS (DATASET_ID) DEFERRABLE INITIALLY DEFERRED,
        PRIMARY KEY (IMAGE_ID, DATASET_ID)
);
CREATE INDEX images_datasets_image_idx ON IMAGE_DATASET_MAP (IMAGE_ID);
CREATE INDEX images_datasets_datasets_idx ON IMAGE_DATASET_MAP (DATASET_ID);
--
--  Datasets are not necessarily contained in a single file.
--  Files do contain contiguous data.
--  For multiple files, how do we put them together into images?
--  Do the files need to know where in the image they belong,
--  or is it up to the images to sort out their files?
--  An OO model might favor that the specific image implementation should worry about its
--  specific file layout.
--  The OME DB schema is supposed to be a fully relational data model, so
--  the file 'object' should know how it fits into the image.
--  This has the additional benefit that now we've standardized file->image mappings in a generic way.
--  At least for 5-D XYZWT images.  Additionally, this is no more painful or messy of an OO model.
--  The assumption with this mapping is that files contain contiguous data, but the individual contiguous blocks
--  can be distributed among an arbitrary set of files, including the pathological condition of one pixel per file (!).
CREATE TABLE IMAGE_FILES_XYZWT (
        IMAGE_ID        OID REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        FILE_SHA1       CHAR (40),     -- SHA1 digest of the file.
        BIGENDIAN       BOOLEAN,       -- wether the file is bigendian or little endian
        PATH            VARCHAR(256),
        HOST            VARCHAR(256),
        URL             VARCHAR(256),
        X_START         SMALLINT,      --  XYZ are pixel coordinates
        X_STOP          SMALLINT,
        Y_START         SMALLINT,
        Y_STOP          SMALLINT,
        Z_START         SMALLINT,
        Z_STOP          SMALLINT,
        W_START         SMALLINT,      --  W is wave number
        W_STOP          SMALLINT,
        T_START         SMALLINT,      --  T is time point
        T_STOP          SMALLINT
);
--
--  This table specifies the image types that OME knows about and how they are implemented.
CREATE SEQUENCE IMAGE_TYPE_SEQ;
CREATE TABLE IMAGE_TYPES (
        IMAGE_TYPE           OID    DEFAULT nextval ('IMAGE_TYPE_SEQ') PRIMARY KEY,
        TYPE_NAME            VARCHAR(64),
        IMPLEMENTATION_LANG  VARCHAR(32),   --  How the interface is implemented i.e. 'Perl', 'Java', 'C', 'C++', etc.
        CLASSPATH            VARCHAR(255),  -- in the implementation language ('OMEDataset::XYZWT','org.ome.image.xyzwt', etc.
        ATTRIBUTES_TABLE     VARCHAR(64)
);
--
ALTER TABLE IMAGES ADD CONSTRAINT image_typefk FOREIGN KEY (IMAGE_TYPE)
        REFERENCES IMAGE_TYPES(IMAGE_TYPE)
        DEFERRABLE INITIALLY DEFERRED;
--
-- One of these per image
CREATE TABLE ATTRIBUTES_IMAGE_XYZWT (
        IMAGE_ID        OID NOT NULL   UNIQUE REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        SIZE_X          INTEGER NOT NULL,    -- Number of X pixels, etc.
        SIZE_Y          INTEGER NOT NULL,
        SIZE_Z          INTEGER NOT NULL,
        NUM_WAVES       INTEGER NOT NULL,
        NUM_TIMES       INTEGER NOT NULL,
        BITS_PER_PIXEL  SMALLINT NOT NULL,
        PIXEL_SIZE_X    FLOAT4,     -- These are in microns per pixel.
        PIXEL_SIZE_Y    FLOAT4,
        PIXEL_SIZE_Z    FLOAT4,
        WAVE_INCREMENT  FLOAT4,     -- This is in nm.
        TIME_INCREMENT  FLOAT4      -- This is in seconds.
);
        -- The following index is not necessary; unique columns are automatically indexed.
-- CREATE INDEX ATTRIBUTES_IMAGE_XYZWT_ID_idx ON ATTRIBUTES_IMAGE_XYZWT (IMAGE_ID);
--
-- One of these per XYZ raster (one per time point per wavelength per image)
CREATE TABLE XYZ_IMAGE_INFO (
        IMAGE_ID      OID      REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        WAVENUMBER      INTEGER,    -- This is the order of the wavelengths in the image.
        TIMEPOINT       INTEGER,    -- This is a timepoint #
        DELTATIME       FLOAT,      -- This is time in seconds relative to the first XYZ.
                                                                -- More precisely, this is the deltatime of the first Z section in the stack
        MIN             INTEGER,
        MAX             INTEGER,
        MEAN            FLOAT4,
        GEOMEAN         FLOAT4,
        SIGMA           FLOAT4,
        CENTROID_X      FLOAT4,     -- These are the "center of mass" for the XYZ stack.
        CENTROID_Y      FLOAT4,
        CENTROID_Z      FLOAT4
);
CREATE INDEX XYZ_INFO_ID_idx ON XYZ_IMAGE_INFO (IMAGE_ID);
--
-- One of these per XY-plane in an XYZWT stack.
CREATE TABLE XY_IMAGE_INFO (
        IMAGE_ID      OID      REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        WAVENUMBER      INTEGER,    -- This is the order of the wavelengths in the image.
        TIMEPOINT       INTEGER,    -- This is a timepoint #
        ZSECTION        INTEGER,    -- The Z section number for these stats
        DELTATIME       FLOAT,      -- This is time in seconds relative to the first plane acquired for this image.
        EXPTIME         FLOAT,      -- The exposure time in seconds
        STAGE_X         FLOAT,      -- Stage coordinates in microscope's reference frame.  FIXME:  units???.
        STAGE_Y         FLOAT,      -- only meaningful relative to each other or with knowledge of what the microscope frame is.
        STAGE_Z         FLOAT,
        MIN             INTEGER,
        MAX             INTEGER,
        MEAN            FLOAT4,
        GEOMEAN         FLOAT4,
        SIGMA           FLOAT4
);
CREATE INDEX XY_INFO_ID_idx ON XY_IMAGE_INFO (IMAGE_ID);
--
-- One of these per wavelength in the image.
CREATE TABLE IMAGE_WAVELENGTHS (
        IMAGE_ID      OID      REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        WAVENUMBER      INTEGER,    -- This is the order of the wavelengths in the image.
        EX_WAVELENGTH   INTEGER,    -- This is in nm.
        EM_WAVELENGTH   INTEGER,    -- This is in nm.
        ND_FILTER       FLOAT,      -- This is a percentage between 0.0 and 1.0.
        FLUOR           VARCHAR(32) -- This is the fluorescent compound (i.e. GFP,FITC, TR, etc.  Not GFP-NFAT or TR-goat-anti-rabbit).
);
CREATE INDEX IMAGE_WAVELENGTHS_ID_idx ON IMAGE_WAVELENGTHS (IMAGE_ID);
--
-- This table contains SoftWorx-specific data for each XY plane.
CREATE TABLE XY_softworx_info (
        IMAGE_ID           OID      REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        WAVENUMBER           INTEGER,    -- This is the order of the wavelengths in the image.
        TIMEPOINT            INTEGER,    -- This is a timepoint #
        ZSECTION             INTEGER,    -- The Z section number for these stats
        PHOTOSENSOR_READING  FLOAT,
        INTEN_SCALING        FLOAT
);
CREATE INDEX XY_softworx_info_ID_idx ON XY_softworx_info (IMAGE_ID);
--
CREATE SEQUENCE SCREEN_SEQ;
CREATE TABLE SCREENS (
        SCREEN_ID        OID    DEFAULT nextval ('SCREEN_SEQ') PRIMARY KEY,
        NAME             VARCHAR(255),
        DESCRIPTION      TEXT,
        EXTERN_SCREEN_ID varchar(40)  --  This refers to the site's screen/plate database.
);
--
CREATE SEQUENCE PLATE_SEQ;
CREATE TABLE PLATES (
        PLATE_ID        OID    DEFAULT nextval ('PLATE_SEQ') PRIMARY KEY,
        NAME            VARCHAR(255),
        EXTERN_PLATE_ID varchar(40)   --  This refers to the site's screen/plate database.
);
--  Establishes a many-to-many relationship between plates and screens
--  This means plates can be re-cycled for different screens.
CREATE TABLE SCREEN_PLATE_MAP (
        SCREEN_ID       OID    REFERENCES SCREENS DEFERRABLE INITIALLY DEFERRED,
        PLATE_ID        OID    REFERENCES PLATES DEFERRABLE INITIALLY DEFERRED,
        PRIMARY KEY (SCREEN_ID, PLATE_ID)
);
--
--  One of these per Dataset that is part of a 'screen'.
--  Note the unique constraint on IMAGE_ID, meaning that a image can belong to only one plate.
CREATE TABLE IMAGE_SCREEN_INFO (
        IMAGE_ID        OID    UNIQUE REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        PLATE_ID          OID    REFERENCES PLATES DEFERRABLE INITIALLY DEFERRED,
        WELL              VARCHAR(8),
        SAMPLE            INTEGER,    -- This is the sample number within a well.
        EXTERN_SAMPLE_ID  varchar(40) --  This refers to the site's screen/plate database (usually what's in the well).
);
--
--  One of these per Dataset to contain a 'stage' label.
--  Note the unique constraint on IMAGE_ID, meaning that a image can belong to only one plate.
CREATE TABLE IMAGE_STAGE_INFO (
        IMAGE_ID        OID    UNIQUE REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
        LABEL             TEXT,     -- A text 'label' for this stage position.
        STAGE_X         FLOAT,      -- Stage coordinates in microscope's reference frame.  FIXME:  units???.
        STAGE_Y         FLOAT,      -- Since there is only one of these per image, the convention is that this is the
        STAGE_Z         FLOAT       -- average stage position for the entire image.
);                          -- Stage positions for individual planes are store in xy_image_info.
--
--  This is now a subclass of XYZWT (as of OME-1.3-IR5)
--CREATE TABLE ATTRIBUTES_ICCB_TIF (
--IMAGE_ID      OID       REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
--SIZE_X          INTEGER,      -- in ATTRIBUTES_IMAGE_XYZWT
--SIZE_Y          INTEGER,      -- ditto
--NUM_WAVES       INTEGER,      -- ibid
--MIN             INTEGER,      -- in XY_IMAGE_INFO
--MAX             INTEGER,      -- ditto
--MEAN            FLOAT,        -- ibid
--SIGMA           FLOAT,        -- metoo
--BASE_NAME       VARCHAR(64),  -- in IMAGES.NAME
--CHEM_PLATE      VARCHAR(64),  -- moved to PLATE.NAME
--COMPOUND_ID     OID,          -- referenced by IMAGE_SCREEN_INFO.EXTERN_SAMPLE_ID
--WAVE            INTEGER,      -- in XY_IMAGE_INFO
--WELL            CHAR(3),      -- in IMAGE_SCREEN_INFO.WELL
--SAMPLE          INTEGER,      -- in IMAGE_SCREEN_INFO.SAMPLE
--RASTER_ID       OID           -- no longer used
--);
--CREATE INDEX ICCB_TIFF_RASTER_idx ON ATTRIBUTES_ICCB_TIFF (RASTER_ID);
--CREATE INDEX ICCB_TIFF_ID_idx ON ATTRIBUTES_ICCB_TIFF (IMAGE_ID);
--
--  This is also no longer used as of OME-1.3-IR5  This data goes into XY_IMAGE_INFO
--CREATE TABLE ATTRIBUTES_TIFF (
--IMAGE_ID      OID       REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
--SIZE_X          INTEGER,
--SIZE_Y          INTEGER,
--MIN             INTEGER,
--MAX             INTEGER,
--MEAN            FLOAT,
--SIGMA           FLOAT
--);
--CREATE INDEX TIFF_ID_idx ON ATTRIBUTES_TIFF (IMAGE_ID);

--
-- Feature attributes list - consult the UML diagram for the implementation.
-- Basically,  an ATTRIBUTE_LIST_ID is a LIST_ID in the ATTRIBUTE_LIST table.  The ATTRIBUTE_LIST table
-- is a map between LIST_ID and table names.  All TABLEs in the ATTRIBUTE_LIST table with the same
-- LIST_ID are part of a given ATTRIBUTE_LIST_ID.  This lets us mix and match feature attributes.  Also
-- it is easy to create user attributes and add them to features.
--
-- The feature itself has no ATTRIBUTE_LIST_ID.  The full list of attributes for a feature can only come from
-- an analysis (ANALYSES table), which has a reference to the PROGRAMS table (threough PROGRAM_ID), 
-- which in turn has a list of attributes.
--
-- Programs can add attributes to existing features (see below).  This is
-- possible because each feature attribute has both a feature id (ATTRIBUTE_OD) and an ANALYSIS_ID.
-- The FEATURES table is a map (many to many relationship) between ANALYSIS and FEATURE.  This
-- makes it possible to have more than one analysis associated with a given feature, and threrefore a given feature
-- can have more than one set of attributes.  One problem arrises if we want to list all attributes for a given
-- feature.  If we know the attributes we want, we can simply join to the FEATURES table instead of the ANALYSES table.
-- If we want to find out what attributes a given feature has, its a little trickier.  We can either find all the table
-- names listed in the ATTRIBUTE_LIST and join them to the feature through the ANALYSES and PROGRAMS table, or we can
-- look through all possible attribute tables for the feature id.
-- As of Version 0.3, each feature attribute has a reference to the analysis that created it (ANALYSIS_ID) along with
-- a reference to the feature it belongs to (ATTRIBUTE_OF).  The ANALYSIS_ID was added because I decided that
-- attributes can be computed for a given feature by more than one analysis.  i.e., analysis 2 adds entries to the 
-- FEATURES table, using its ANALYSIS_ID, but an existing FEATURE_ID (i.e. from analysis 1).
-- Without the ANALYSIS_ID, a given attribute entry can be generated by more than one analysis, but still have the same
-- ATTRIBUTE_OF.  This would mean that we could have rows in the attribute tables that are indistinguishable.  A bad thing.
-- Even though the rows are still distinguisable because every row has a system-assigned OID, this is not always the case,
-- and we don't rely on this system OID anywhere else, making them explicit instead.
--

--
-- 24 September 2002 (DC)
-- Some new tables for the analysis subsystem

CREATE SEQUENCE LOOKUP_TABLE_SEQ;
CREATE TABLE LOOKUP_TABLES (
  LOOKUP_TABLE_ID  INTEGER DEFAULT nextval('LOOKUP_TABLE_SEQ') PRIMARY KEY,
  NAME             VARCHAR(64) NOT NULL,
  DESCRIPTION      TEXT
);
CREATE INDEX LOOKUP_TABLES_NAME ON LOOKUP_TABLES(NAME);

CREATE SEQUENCE LOOKUP_TABLE_ENTRY_SEQ;
CREATE TABLE LOOKUP_TABLE_ENTRIES (
  LOOKUP_TABLE_ENTRY_ID  INTEGER DEFAULT nextval('LOOKUP_TABLE_ENTRY_SEQ') PRIMARY KEY,
  LOOKUP_TABLE_ID        INTEGER NOT NULL REFERENCES LOOKUP_TABLES DEFERRABLE INITIALLY DEFERRED,
  VALUE                  VARCHAR(256) NOT NULL,
  LABEL                  VARCHAR(256)
);
CREATE INDEX LOOKUP_TABLE_ENTRIES_LT_ID ON LOOKUP_TABLE_ENTRIES(LOOKUP_TABLE_ID);

CREATE SEQUENCE DATATYPE_SEQ;
CREATE TABLE DATATYPES (
  DATATYPE_ID  INTEGER DEFAULT nextval('DATATYPE_SEQ') PRIMARY KEY,
  TABLE_NAME   VARCHAR(64) NOT NULL,
  DESCRIPTION  TEXT
);
CREATE INDEX DATATYPES_TABLE_NAME ON DATATYPES(TABLE_NAME);

CREATE SEQUENCE DATATYPE_COLUMN_SEQ;
CREATE TABLE DATATYPE_COLUMNS (
  DATATYPE_COLUMN_ID   INTEGER DEFAULT nextval('DATATYPE_COLUMN_SEQ') PRIMARY KEY,
  DATATYPE_ID          INTEGER NOT NULL REFERENCES DATATYPES DEFERRABLE INITIALLY DEFERRED,
  COLUMN_NAME          VARCHAR(256) NOT NULL,
  REFERENCE_TYPE       VARCHAR(256)
);
CREATE INDEX DATATYPE_COLUMNS_DATATYPE_ID ON DATATYPE_COLUMNS(DATATYPE_ID);
CREATE INDEX DATATYPE_COLUMNS_COLUMN_NAME ON DATATYPE_COLUMNS(COLUMN_NAME);

--
-- *****
-- 17 May 2002 (DC)
--   - Changed PROGRAMS table to require analyses to implement a Perl class (specified by the LOCATION field).
--   - Eventually, this will migrate to an XML-RPC/SOAP/etc. implementation, and the LOCATION field
--     will be interpreted differently (as a URL, most likely)
--
-- 24 September 2002
--   - Removed INPUT_TABLE column since this is now specified in
--     the formal inputs table
CREATE SEQUENCE PROGRAM_SEQ;
CREATE TABLE PROGRAMS (
       PROGRAM_ID     OID DEFAULT NEXTVAL('PROGRAM_SEQ') PRIMARY KEY,
       PROGRAM_NAME   VARCHAR(64) NOT NULL,
       DESCRIPTION    TEXT,
       LOCATION       VARCHAR(128) NOT NULL,
       MODULE_TYPE    VARCHAR(128) NOT NULL,
       CATEGORY       VARCHAR(32) NOT NULL,
       VISUAL_DESIGN  TEXT
);
CREATE INDEX PROGRAMS_PROGRAM_NAME ON PROGRAMS(PROGRAM_NAME);
CREATE INDEX PROGRAMS_CATEGORY ON PROGRAMS(CATEGORY);
CREATE INDEX PROGRAMS_LOCATION ON PROGRAMS(LOCATION);
CREATE INDEX PROGRAMS_MODULE_TYPE ON PROGRAMS(MODULE_TYPE);

-- 24 September 2002 (DC)
-- New tables to store the defined inputs/outputs of the analysis modules.
CREATE SEQUENCE FORMAL_INPUT_SEQ;
CREATE TABLE FORMAL_INPUTS (
  FORMAL_INPUT_ID  INTEGER DEFAULT nextval('FORMAL_INPUT_SEQ') PRIMARY KEY,
  PROGRAM_ID       INTEGER NOT NULL REFERENCES PROGRAMS DEFERRABLE INITIALLY DEFERRED,
  NAME             VARCHAR(64) NOT NULL,
  COLUMN_TYPE      INTEGER NOT NULL REFERENCES DATATYPE_COLUMNS DEFERRABLE INITIALLY DEFERRED,
  LOOKUP_TABLE_ID  INTEGER REFERENCES LOOKUP_TABLES
);
CREATE INDEX FORMAL_INPUTS_PROGRAM_ID ON FORMAL_INPUTS(PROGRAM_ID);
CREATE INDEX FORMAL_INPUTS_NAME ON FORMAL_INPUTS(NAME);
CREATE INDEX FORMAL_INPUTS_COLUMN_TYPE ON FORMAL_INPUTS(COLUMN_TYPE);

CREATE SEQUENCE FORMAL_OUTPUT_SEQ;
CREATE TABLE FORMAL_OUTPUTS (
  FORMAL_OUTPUT_ID  INTEGER DEFAULT nextval('FORMAL_OUTPUT_SEQ') PRIMARY KEY,
  PROGRAM_ID        INTEGER NOT NULL REFERENCES PROGRAMS DEFERRABLE INITIALLY DEFERRED,
  NAME              VARCHAR(64) NOT NULL,
  COLUMN_TYPE       INTEGER NOT NULL REFERENCES DATATYPE_COLUMNS DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX FORMAL_OUTPUTS_PROGRAM_ID ON FORMAL_OUTPUTS(PROGRAM_ID);
CREATE INDEX FORMAL_OUTPUTS_NAME ON FORMAL_OUTPUTS(NAME);
CREATE INDEX FORMAL_OUTPUTS_COLUMN_TYPE ON FORMAL_OUTPUTS(COLUMN_TYPE);


-- analysis views (a particular graph of modules linked via inputs/outputs)

CREATE SEQUENCE ANALYSIS_VIEW_SEQ;
CREATE TABLE ANALYSIS_VIEWS (
  ANALYSIS_VIEW_ID  OID DEFAULT nextval('ANALYSIS_VIEW_SEQ') PRIMARY KEY,
  OWNER             OID NOT NULL REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
  NAME              VARCHAR(64) NOT NULL
);
CREATE INDEX ANALYSIS_VIEWS_OWNER ON ANALYSIS_VIEWS(OWNER);

CREATE SEQUENCE ANALYSIS_VIEW_NODES_SEQ;
CREATE TABLE ANALYSIS_VIEW_NODES (
  ANALYSIS_VIEW_NODE_ID  OID DEFAULT nextval('ANALYSIS_VIEW_NODES_SEQ') PRIMARY KEY, 
  ANALYSIS_VIEW_ID       OID NOT NULL REFERENCES ANALYSIS_VIEWS DEFERRABLE INITIALLY DEFERRED,
  PROGRAM_ID             OID NOT NULL REFERENCES PROGRAMS DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX ANALYSIS_VIEW_NODES_VIEW ON ANALYSIS_VIEW_NODES(ANALYSIS_VIEW_ID);
CREATE INDEX ANALYSIS_VIEW_NODES_PROGRAM ON ANALYSIS_VIEW_NODES(PROGRAM_ID);
 
CREATE SEQUENCE ANALYSIS_VIEW_LINKS_SEQ;
CREATE TABLE ANALYSIS_VIEW_LINKS (
  ANALYSIS_VIEW_LINK_ID  OID DEFAULT nextval('ANALYSIS_VIEW_LINKS_SEQ') PRIMARY KEY,
  ANALYSIS_VIEW_ID       OID NOT NULL REFERENCES ANALYSIS_VIEWS DEFERRABLE INITIALLY DEFERRED,
  FROM_NODE              OID NOT NULL REFERENCES ANALYSIS_VIEW_NODES DEFERRABLE INITIALLY DEFERRED,
  FROM_0UTPUT            OID NOT NULL REFERENCES FORMAL_OUTPUTS DEFERRABLE INITIALLY DEFERRED,
  TO_NODE                OID NOT NULL REFERENCES ANALYSIS_VIEW_NODES DEFERRABLE INITIALLY DEFERRED,
  TO_INPUT               OID NOT NULL REFERENCES FORMAL_INPUTS DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX ANALYSIS_VIEW_LINKS_VIEW ON ANALYSIS_VIEW_LINKS(ANALYSIS_VIEW_ID);
CREATE INDEX ANALYSIS_VIEW_LINKS_FROM_NODE ON ANALYSIS_VIEW_LINKS(FROM_NODE);
CREATE INDEX ANALYSIS_VIEW_LINKS_FROM_OUTPUT ON ANALYSIS_VIEW_LINKS(FROM_OUTPUT);
CREATE INDEX ANALYSIS_VIEW_LINKS_TO_NODE ON ANALYSIS_VIEW_LINKS(TO_NODE);
CREATE INDEX ANALYSIS_VIEW_LINKS_TO_OUTPUT ON ANALYSIS_VIEW_LINKS(TO_INPUT);

--
CREATE SEQUENCE ANALYSIS_SEQ;
CREATE TABLE ANALYSES (
        ANALYSIS_ID     OID       DEFAULT nextval ('ANALYSIS_SEQ') PRIMARY KEY,
        PROGRAM_ID      OID       NOT NULL REFERENCES PROGRAMS DEFERRABLE INITIALLY DEFERRED,
        DATASET_ID      OID       NOT NULL REFERENCES DATASETS DEFERRABLE INITIALLY DEFERRED,
        EXPERIMENTER_ID OID       NOT NULL REFERENCES EXPERIMENTERS DEFERRABLE INITIALLY DEFERRED,
        TIMESTAMP       TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
        STATUS          VARCHAR(16)
);
CREATE INDEX ANALYSES_PROGRAM_ID ON ANALYSES(PROGRAM_ID);
CREATE INDEX ANALYSES_EXPERIMENTER_ID ON ANALYSES(EXPERIMENTER_ID);
CREATE INDEX ANALYSES_DATASET_ID ON ANALYSES(DATASET_ID);

CREATE TABLE ANALYSIS_PATH_MAP (
  PATH_ID      OID,
  ANALYSIS_ID  OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
  PRIMARY KEY (PATH_ID, ANALYSIS_ID)
);
CREATE INDEX ANALYSIS_PATH_PATH ON ANALYSIS_PATH_MAP(PATH_ID);
CREATE INDEX ANALYSIS_PATH_ANALYSIS ON ANALYSIS_PATH_MAP(ANALYSIS_ID);

CREATE SEQUENCE ACTUAL_INPUT_SEQ;
CREATE TABLE ACTUAL_INPUTS (
  ACTUAL_INPUT_ID  INTEGER DEFAULT nextval('ACTUAL_INPUT_SEQ') PRIMARY KEY,
  ANALYSIS_ID      INTEGER NOT NULL REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
  FORMAL_INPUT_ID  INTEGER NOT NULL REFERENCES FORMAL_INPUTS DEFERRABLE INITIALLY DEFERRED,
  IMAGE_ID         INTEGER NOT NULL REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
  ATTRIBUTE_ID     INTEGER NOT NULL,
  UNIQUE (ANALYSIS_ID, FORMAL_INPUT_ID, IMAGE_ID)
);
CREATE INDEX ACTUAL_INPUTS_ANALYSIS_ID ON ACTUAL_INPUTS(ANALYSIS_ID);
CREATE INDEX ACTUAL_INPUTS_FI_ID ON ACTUAL_INPUTS(FORMAL_INPUT_ID);
CREATE INDEX ACTUAL_INPUTS_IMAGE_ID ON ACTUAL_INPUTS(IMAGE_ID);
CREATE INDEX ACTUAL_INPUTS_ATTRIBUTE_ID ON ACTUAL_INPUTS(ATTRIBUTE_ID);

CREATE SEQUENCE ACTUAL_OUTPUT_SEQ;
CREATE TABLE ACTUAL_OUTPUTS (
  ACTUAL_OUTPUT_ID  INTEGER DEFAULT nextval('ACTUAL_OUTPUT_SEQ') PRIMARY KEY,
  ANALYSIS_ID       INTEGER NOT NULL REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
  FORMAL_OUTPUT_ID  INTEGER NOT NULL REFERENCES FORMAL_OUTPUTS DEFERRABLE INITIALLY DEFERRED,
  IMAGE_ID          INTEGER NOT NULL REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
  ATTRIBUTE_ID      INTEGER NOT NULL,
  UNIQUE (ANALYSIS_ID, FORMAL_OUTPUT_ID, IMAGE_ID)
);
CREATE INDEX ACTUAL_OUTPUTS_ANALYSIS_ID ON ACTUAL_OUTPUTS(ANALYSIS_ID);
CREATE INDEX ACTUAL_OUTPUTS_FO_ID ON ACTUAL_OUTPUTS(FORMAL_OUTPUT_ID);
CREATE INDEX ACTUAL_OUTPUTS_IMAGE_ID ON ACTUAL_OUTPUTS(IMAGE_ID);
CREATE INDEX ACTUAL_OUTPUTS_ATTRIBUTE_ID ON ACTUAL_OUTPUTS(ATTRIBUTE_ID);

CREATE SEQUENCE ATTRIBUTE_SEQ;

-- "My First Analysis Module"
-- 25 September 2002

CREATE TABLE IMAGE_STATISTICS (
  ATTRIBUTE_ID  INTEGER DEFAULT nextval('ATTRIBUTE_SEQ') PRIMARY KEY,
  MEAN          FLOAT,
  SIGMA         FLOAT,
  CENTROID_X    INTEGER,
  CENTROID_Y    INTEGER,
  CENTROID_Z    INTEGER,
  CENTROID_W    INTEGER,
  CENTROID_T    INTEGER
);

INSERT INTO DATATYPES (DATATYPE_ID,TABLE_NAME,DESCRIPTION)
VALUES (NEXTVAL('DATATYPE_SEQ'),'IMAGE_STATISTICS','Image-wide statistics');

INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'MEAN',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'SIGMA',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'CENTROID_X',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'CENTROID_Y',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'CENTROID_Z',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'CENTROID_W',NULL);
INSERT INTO DATATYPE_COLUMNS (DATATYPE_COLUMN_ID,DATATYPE_ID,COLUMN_NAME,REFERENCE_TYPE)
VALUES (NEXTVAL('DATATYPE_COLUMN_SEQ'),CURRVAL('DATATYPE_SEQ'),'CENTROID_T',NULL);

--
--
-- The FEATURES table is used for ALL image features reguardless of algorithm.
-- The properties of a feature are that it has a variable list of attributes.
-- There are attributes specified by OME, with a mechanism for making user attributes (see
-- documentation above PROGRAMS)
--  The FEATURES table has a feature id as a primary key, and the associated ANALYSIS_ID is the
-- analysis that created the feature.  The ATTRIBUTES table has both feature and analysis IDs as foreign
-- keys - this table is a map between analyses and features in a many to many relationship.
-- This table lists all analyses that produced attributes for a given feature.
CREATE SEQUENCE FEATURE_SEQ; 
CREATE TABLE FEATURES (
        FEATURE_ID     OID DEFAULT nextval ('FEATURE_SEQ') PRIMARY KEY,
        ANALYSIS_ID    OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX features_analysis_id_idx ON features (analysis_id);
--
-- FIXME:  Is this actually used for anything?
CREATE TABLE ATTRIBUTES (
        FEATURE_ID     OID REFERENCES FEATURES DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID    OID REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED
);
CREATE INDEX attr_feature_id_idx ON attributes (feature_id);
CREATE INDEX attr_analysis_id_idx ON attributes (analysis_id);
--
-- This table groups table names by LIST_ID
-- It seems like we could just refer to PROGRAM_ID rather than going through LIST_ID.
-- Can we have more than one program with the same LIST_ID?  In practice, probably not.
-- Can we have more than one list associated with a program?  No.
-- In Version 1.3, removed the redundant LIST_ID, and refered to PROGRAM_ID instead.
CREATE TABLE ATTRIBUTE_LIST (
        PROGRAM_ID        OID          REFERENCES PROGRAMS DEFERRABLE INITIALLY DEFERRED,
        TABLE_NAME        VARCHAR(64)
);
--
-- This table maintains a list of expired files - files to be deleted from the file system
-- as one of the periodic database maintenance chores.
-- Entries in this table should be deleted after the files are actually removed from the system.
-- Attributes that maintain data in files on the filesystem should create rules such as:
-- CREATE RULE expire_mytable_files AS
--   ON DELETE TO mytable DO
--   INSERT INTO expired_files VALUES (OLD.path || OLD.name);
CREATE TABLE expired_files (
  fullpath    varchar(255)
);
--
--
-- This is a table with project and image view templates.
-- variable substitution will occur wherever template matches $(ProjectID) or $(DatasetID)
-- The type is either 'Dataset' or 'Project'.  Possibly other types to follow.
CREATE TABLE view_templates (
        name          varchar(32),
        description   text,
        template      text,
        type          varchar(32),
        owner_id         OID REFERENCES experimenters DEFERRABLE INITIALLY DEFERRED,
        group_id         OID REFERENCES groups DEFERRABLE INITIALLY DEFERRED
);
--
--
--
-- What follows is a bunch of tables for presently defined attributes.
-- We don't make attribute_of a foreign key because there are potentially duplicate feature_ids
-- with different analysis_ids in the features table.  For this reason, the feature_id isn't
-- a primary key in the features table.
-- The analysis_id is a primary key in the analyses table, and a foreign key in these tables.
CREATE TABLE LOCATION (  -- Sorry, but POSITION is a reserved word in SQL, apparently.
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        X             FLOAT,
        Y             FLOAT,
        Z             FLOAT
);
CREATE INDEX LOCATION_ATTRIBUTE_OF_idx ON LOCATION (ATTRIBUTE_OF);
CREATE INDEX LOCATION_ANALYSIS_ID_idx  ON LOCATION (ANALYSIS_ID);
--
CREATE TABLE SIGNAL (
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        WAVELENGTH       SMALLINT,
        CENTROID_X      FLOAT,
        CENTROID_Y      FLOAT,
        CENTROID_Z      FLOAT,
        INTEGRAL        FLOAT,
        MEAN            FLOAT,
        GEOMEAN         FLOAT,
        SIGMA           FLOAT,
        SD_OVER_MEAN    FLOAT,
        SD_OVER_GEOMEAN FLOAT,
        BACKGROUND      FLOAT
);
CREATE INDEX SIGNAL_ATTRIBUTE_OF_idx ON SIGNAL (ATTRIBUTE_OF);
CREATE INDEX SIGNAL_ANALYSIS_ID_idx  ON SIGNAL (ANALYSIS_ID);

CREATE TABLE TIMEPOINT (
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        TIMEPOINT     INTEGER
);
CREATE INDEX TIMEPOINT_ATTRIBUTE_OF_idx ON TIMEPOINT (ATTRIBUTE_OF);
CREATE INDEX TIMEPOINT_ANALYSIS_ID_idx  ON TIMEPOINT (ANALYSIS_ID);

CREATE SEQUENCE TRAJECTORY_ID_SEQ;
CREATE TABLE DELTA (
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        MOVED_TO      OID,
        DELTA_X       FLOAT,
        DELTA_Y       FLOAT,
        DELTA_Z       FLOAT,
        DISTANCE      FLOAT,
        VELOCITY      FLOAT,
        TRAJECTORY_ID OID
);
CREATE INDEX DELTA_ATTRIBUTE_OF_idx ON DELTA (ATTRIBUTE_OF);
CREATE INDEX DELTA_ANALYSIS_ID_idx  ON DELTA (ANALYSIS_ID);

--
CREATE TABLE EXTENT (
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        VOLUME        INTEGER,
        MIN_X         SMALLINT,
        MIN_Y         SMALLINT,
        MIN_Z         SMALLINT,
        MAX_X         SMALLINT,
        MAX_Y         SMALLINT,
        MAX_Z         SMALLINT,
        SIGMA_X       FLOAT,
        SIGMA_Y       FLOAT,
        SIGMA_Z       FLOAT,
        SURFACE_AREA  FLOAT,    -- For 2D spots, this is the surface area including fenestrations.
        PERIMITER     FLOAT,    -- The outside perimiter - this will not include inside perimiters in fenestrated spots
        FORM_FACTOR   FLOAT     -- surface area/volume (3D) or perimiter/surface area (2D) 1.0 for circle or sphere.
);
CREATE INDEX EXTENT_ATTRIBUTE_OF_idx ON EXTENT (ATTRIBUTE_OF);
CREATE INDEX EXTENT_ANALYSIS_ID_idx  ON EXTENT (ANALYSIS_ID);

CREATE TABLE THRESHOLD (
        ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
        ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
        THRESHOLD     SMALLINT
);
CREATE INDEX THRESHOLD_ATTRIBUTE_OF_idx ON THRESHOLD (ATTRIBUTE_OF);
CREATE INDEX THRESHOLD_ANALYSIS_ID_idx  ON THRESHOLD (ANALYSIS_ID);

CREATE TABLE ENVELOPE (
ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
X             SMALLINT,
Y             SMALLINT,
Z             SMALLINT
);
CREATE INDEX ENVELOPE_ATTRIBUTE_OF_idx ON ENVELOPE (ATTRIBUTE_OF);
CREATE INDEX ENVELOPE_ANALYSIS_ID_idx  ON ENVELOPE (ANALYSIS_ID);

CREATE TABLE DISTANCE_TO_SPINDLE_AXIS (
ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
DISTANCE      FLOAT,
LENGTH        FLOAT
);
CREATE INDEX DISTANCE_TO_SPINDLE_AXIS_ATTRIBUTE_OF_idx ON DISTANCE_TO_SPINDLE_AXIS (ATTRIBUTE_OF);
CREATE INDEX DISTANCE_TO_SPINDLE_AXIS_ANALYSIS_ID_idx  ON DISTANCE_TO_SPINDLE_AXIS (ANALYSIS_ID);

CREATE SEQUENCE CLUSTER_ID;
CREATE TABLE CLUSTER_F (
ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
CLUSTER_ID    OID
);
CREATE INDEX CLUSTER_F_ATTRIBUTE_OF_idx ON CLUSTER_F (ATTRIBUTE_OF);
CREATE INDEX CLUSTER_F_ANALYSIS_ID_idx  ON CLUSTER_F (ANALYSIS_ID);

CREATE TABLE NEAREST_NEIGHBOR (
ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
ANALYSIS_ID   OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
FEATURE_ID    OID,
DELTA_X       FLOAT,
DELTA_Y       FLOAT,
DELTA_Z       FLOAT,
DISTANCE      FLOAT
);
CREATE INDEX NEAREST_NEIGHBOR_ATTRIBUTE_OF_idx ON NEAREST_NEIGHBOR (ATTRIBUTE_OF);
CREATE INDEX NEAREST_NEIGHBOR_ANALYSIS_ID_idx  ON NEAREST_NEIGHBOR (ANALYSIS_ID);

CREATE TABLE BINARY_IMAGE (
ATTRIBUTE_OF  OID        REFERENCES FEATURES (FEATURE_ID) DEFERRABLE INITIALLY DEFERRED,
ANALYSIS_ID         OID        REFERENCES ANALYSES (ANALYSIS_ID) DEFERRABLE INITIALLY DEFERRED,
image_type        VARCHAR(32),     -- Should match one of the IMAGE_TYPEs in the IMAGES table.
image_id_in       OID,             -- The image id that this is a binary image of.
image_id_bin      OID,             -- Does not necessarily point to an actual image ID (could be NULL).
name                VARCHAR(256),
path                VARCHAR(256)
);
CREATE INDEX BINARY_IMAGE_ATTRIBUTE_OF_idx ON BINARY_IMAGE (ATTRIBUTE_OF);
CREATE INDEX BINARY_IMAGE_ANALYSIS_ID_idx  ON BINARY_IMAGE (ANALYSIS_ID);
CREATE INDEX BINARY_IMAGE_IMAGE_ID_idx  ON BINARY_IMAGE (image_id_in);
CREATE RULE expire_BINARY_IMAGE_files AS
  ON DELETE TO BINARY_IMAGE DO
  INSERT INTO expired_files VALUES (OLD.path || OLD.name);

--
--
--
--  Declaration of the programs.
CREATE TABLE INPUT_FINDSPOTS (
ANALYSIS_ID    OID      REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
TIME_START     INTEGER,
TIME_STOP      INTEGER,
WAVELENGTH     INTEGER,
THRESHOLD      VARCHAR(16),
MIN_SPOT_VOL   INTEGER
);
--
INSERT INTO PROGRAMS (PROGRAM_ID,PROGRAM_NAME,LOCATION,DESCRIPTION,CATEGORY,INPUT_TABLE) VALUES (
nextval ('PROGRAM_SEQ'),
'findSpots',
'OMEAnalysis::FindSpots',
'Simple threshold segmentation',
'Segmentation',
'INPUT_FINDSPOTS'
);
--
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'THRESHOLD');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'EXTENT');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'TIMEPOINT');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'SIGNAL');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'LOCATION');
--
--
CREATE TABLE INPUT_TRACKSPOTS (
ANALYSIS_ID    OID      REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
TIME_START     INTEGER,
TIME_STOP      INTEGER,
WAVELENGTH     INTEGER,
THRESHOLD      VARCHAR(16),
MIN_SPOT_VOL   INTEGER,
INTENS_WEIGHT  FLOAT
);
--
-- Note that an absolute path to the binary must also be specified (in attribute 'path')
INSERT INTO PROGRAMS (PROGRAM_ID,PROGRAM_NAME,LOCATION,DESCRIPTION,CATEGORY,INPUT_TABLE) VALUES (
nextval ('PROGRAM_SEQ'),
'trackSpots',
'OMEAnalysis::TrackSpots',
'Uses segmented data to track features through time',
'Tracking',
'INPUT_TRACKSPOTS'
);
--
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'THRESHOLD');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'EXTENT');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'TIMEPOINT');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'SIGNAL');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'LOCATION');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'DELTA');
--
--
CREATE TABLE INPUT_TRACKSPOTS2 (
ANALYSIS_ID    OID      REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
INTENS_WEIGHT  FLOAT
);
--
-- Note that an absolute path to the binary must also be specified (in attribute 'path')
INSERT INTO PROGRAMS (PROGRAM_ID,PROGRAM_NAME,LOCATION,DESCRIPTION,CATEGORY,INPUT_TABLE) VALUES (
nextval ('PROGRAM_SEQ'),
'trackSpots2',
'OMEAnalysis::TrackSpots2',
'Uses segmented data to track features through time',
'Tracking',
'INPUT_TRACKSPOTS2'
);
--
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'DELTA');
--
--
--
CREATE TABLE INPUT_GETNEARESTNEIGHBORS (
ANALYSIS_ID    OID    REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED
);
--
INSERT INTO PROGRAMS (PROGRAM_ID,PROGRAM_NAME,LOCATION,DESCRIPTION,CATEGORY,INPUT_TABLE) VALUES (
nextval ('PROGRAM_SEQ'),
'getNearestNeighbors',
'OMEAnalysis::NearestNeighbors',
'Calculates a nearest neighbor for each feature',
'Clustering',
'INPUT_GETNEARESTNEIGHBORS'
);
--
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'NEAREST_NEIGHBOR');
--
--
--
CREATE TABLE INPUT_BINARIZE_GLOBAL (
ANALYSIS_ID    OID    REFERENCES ANALYSES DEFERRABLE INITIALLY DEFERRED,
METHOD         VARCHAR(32),
ARGUMENT       VARCHAR(32),    -- The argument to the thresholding program (if any).
MAKE_BINARY    BOOLEAN,        -- Wether or not binary images were made.
WAVELENGTH     INTEGER         -- The wavelength the binarization program operated on.
);
--
INSERT INTO PROGRAMS (PROGRAM_ID,PROGRAM_NAME,LOCATION,DESCRIPTION,CATEGORY,INPUT_TABLE) VALUES (
nextval ('PROGRAM_SEQ'),
'BinarizeGlobal',
'OMEAnalysis::BinarizeGlobal',
'Create a binary image using a global threshold calculated in one of several different ways.',
'Segmentation',
'INPUT_BINARIZE_GLOBAL'
);
--
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'BINARY_IMAGE');
INSERT INTO ATTRIBUTE_LIST (PROGRAM_ID,TABLE_NAME) VALUES (currval('PROGRAM_SEQ'),'THRESHOLD');
--
--
--
-- Make a rule or trigger to delete old entries where ome_name is the same
-- as newly added ome_name, i.e. one session per user.
-- This is currently done in OMElogin.pl
-- OMElogin.pl preserves a user's settings from a previous session
--
CREATE SEQUENCE SESSION_SEQ; 
CREATE TABLE OME_SESSIONS (
SESSION_ID         OID        DEFAULT nextval ('SESSION_SEQ') PRIMARY KEY,
SESSION_KEY        VARCHAR(36),                 -- The client's state info (not the user's state info, which is SESSION_ID)
OME_NAME           VARCHAR(30),                 -- The database user name
HOST               VARCHAR(30),                 -- The remote host
PROJECT_ID         OID        REFERENCES PROJECTS DEFERRABLE INITIALLY DEFERRED,
IMAGE_VIEW       TEXT,                        -- A select clause ending with where
FEATURE_VIEW       TEXT,                        -- A select clause ending with where - the view name will be the imageName.
ANALYSIS           VARCHAR(64),                 -- A relative URL to the page in the MainFrame - should be a program ID?
DISPLAY_SETTINGS   OID REFERENCES image_display_settings DEFERRABLE INITIALLY DEFERRED,
LAST_ACCESS        TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
STARTED            TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);
CREATE TABLE OME_SESSIONS_IMAGES (
IMAGE_ID         OID REFERENCES IMAGES DEFERRABLE INITIALLY DEFERRED,
SESSION_ID         OID REFERENCES OME_SESSIONS DEFERRABLE INITIALLY DEFERRED
);






