#!/usr/bin/perl -w
#
# sample-populate.pl
# Copyright (C) 2003 Open Microscopy Environment, MIT
# Author:  Brian S. Hughes
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    This script will populate the OME database with sample entries
#    for use in testing and development.

use strict;
use DBI;
use DBD::Pg;

my $dbh;
my $sth;
my $ndx;
my $i;
my $tblname;
my $colnames;

my @tables = 
    ('groups', ['name'], [['yeast'], ['mammalian'], ['screeners']],

     'experimenters', ['ome_name', 'firstname', 'lastname', 'group_id', 'password'], 
     [['dcreager', 'Doug', 'Creager', '@groups.attribute_id.name.OME', '~passwd'],
      ['drrines', 'Dan', 'Rines', '@groups.attribute_id.name.yeast', '~passwd'],
      ['sgaudet', 'Suzannes', 'Gaudet', '@groups.attribute_id.name.yeast', '~passwd'],
      ['aburds', 'Aurora', 'Burds', '@groups.attribute_id.name.mammalian', '~passwd'],
      ['amcainsh', 'Andrew', 'Mcainsh', '@groups.attribute_id.name.yeast', '~passwd'],
      ['rhagan', 'Rob', 'Hagan', '@groups.attribute_id.name.mammalian', '~passwd'],
      ['sxie', 'Steffanie', 'Xie', '@groups.attribute_id.name.yeast', '~passwd'],
      ['jtytell', 'Jessica', 'Tytell', '@groups.attribute_id.name.mammalian', '~passwd']
      ],

     'projects', ['name', 'owner_id', 'group_id'],
     [['Chromosome Segregation', '@experimenters.attribute_id.ome_name.drrines', '@groups.attribute_id.name.yeast'],
      ['Checkpoint Release', '@experimenters.attribute_id.ome_name.sgaudet', '@groups.attribute_id.name.yeast'],
      ['Blastosphere Apoptosis', '@experimenters.attribute_id.ome_name.aburds', '@groups.attribute_id.name.mammalian'],
      ['Apoptotic Cascade', '@experimenters.attribute_id.ome_name.sgaudet', '@groups.attribute_id.name.yeast'],
      ['Actin Morphogenesis', '@experimenters.attribute_id.ome_name.rhagan', '@groups.attribute_id.name.mammalian'],
      ['Kinetechore Complexes', '@experimenters.attribute_id.ome_name.amcainsh', '@groups.attribute_id.name.yeast']],

     'datasets', ['name', 'owner_id', 'group_id', 'description'],
     [['Baseline Spindle Assembly', '@experimenters.attribute_id.ome_name.drrines', '@groups.attribute_id.name.yeast', 'wt spindle assembly'],
      ['Lagging Segregation', '@experimenters.attribute_id.ome_name.drrines', '@groups.attribute_id.name.yeast', 'chromosome segregation at 37C'],
      ['bub1a Mutant Apoptosis', '@experimenters.attribute_id.ome_name.sgaudet', '@groups.attribute_id.name.yeast', 'bub1d apoptosis'],
      ['DAM1 Associations', '@experimenters.attribute_id.ome_name.amcainsh', '@groups.attribute_id.name.yeast', 'DAM1 role in complexes']],

     'project_dataset_map', ['project_id', 'dataset_id'],
     [['@projects.project_id.name.Chromosome Segregation','@datasets.dataset_id.name.Baseline Spindle Assembly'],
      ['@projects.project_id.name.Chromosome Segregation','@datasets.dataset_id.name.Lagging Segregation'],
      ['@projects.project_id.name.Apoptotic Cascade','@datasets.dataset_id.name.bub1a Mutant Apoptosis'],
      ['@projects.project_id.name.Kinetechore Complexes','@datasets.dataset_id.name.DAM1 Associations']]
     );



my $admin = set_to_admin();  # connect to DB as the DB admin
$dbh = DBI->connect("dbi:Pg:dbname=ome", $admin, '' ,{RaiseError=>1});

# fill in the DB tables listed in the 'tables' array, above
$ndx = scalar @tables;
for ($i = 0; $i < $ndx; $i += 3) {
    $tblname = $tables[$i];
    $colnames = $tables[$i+1];
    my @cols = parse($colnames, "");
    my $valueset = $tables[$i+2];
    my $valset;
    for $valset (@$valueset) {
	my @vals = parse($valset, "\'");
	my $statement = "INSERT INTO $tblname (@cols) VALUES (@vals)";
	$dbh->do($statement);
    }
}

$dbh->disconnect;



# Create list of parameters, separated by specified separator
sub parse {
    my $in_ref = shift;
    my $out_sep = shift;
    my $flds = 0;
    my $v;
    my @lst;
    for $v (@$in_ref){
	if ($v =~ /@.*/) {    # Do a SELECT WHERE to, e.g.  get id value from name 
	    $v = get_id_from_name($v);
	}
	elsif ($v =~ /~.*/) { # access field specific value calculation routine
	    $v = calculate_value($v);
	}
	my $vx = "";
	if ($flds++ > 0) {
	    $vx .= ", ";
	}
	$vx .= $out_sep.$v.$out_sep;
	push @lst, $vx;
    }
    return(@lst);    
}


sub get_id_from_name {
    my $dbRef = shift;
    my $sth;
    my @parts;
    my $tbl;
    my $fld;
    my $name;
    my @row;

    $dbRef =~ s/@(.*)/$1/;
    @parts = split /\./, $dbRef;
    $sth = $dbh->prepare("SELECT $parts[1] FROM $parts[0] WHERE $parts[2] = \'$parts[3]\'");
    $sth->execute();
    @row = $sth->fetchrow_array;
    return $row[0];
}


sub calculate_value {
    my $arg = shift;
    my $what = $arg;
    my $password;
    
    $what =~ s/~(.*)/$1/;
    if ($what eq "passwd") {
	$password = " ";
	my $salt = join('',('.','/',0..9,'A'..'Z','a'..'z')[rand 64, rand 64]);
	$password = crypt(" ", $salt);
	$what = $password;
    }

    return $what;
}
	


# Ask user for DB admin username
sub set_to_admin {
    my $adminUser = "";
    my $user = get_uid();
    print "OME administrative user [$user]: ";
    my $adminUserIn = <STDIN>;
    chomp $adminUserIn;
    if ($adminUserIn) {
        $adminUser = $adminUserIn;
    } else {
	$adminUser = $user;
    }
    return $user;
}

# find out current username
sub get_uid {
my @login = getpwuid($>);
	return ( $login[0] );
}

